# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 上午12:49
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from pydantic import ConfigDict, Field, BaseModel

from .discord import BotSetting as DiscordSetting
from .kook import BotSetting as KookSetting
from .slack import BotSetting as SlackSetting
from .telegram import BotSetting as TelegramSetting


class StartSetting(BaseModel):
    """
    平台列表
    """
    discord: bool = Field(default=False)
    kook: bool = Field(default=False)
    slack: bool = Field(default=False)
    telegram: bool = Field(default=False)
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def from_subdir(cls):
        _kwargs = {}
        if SlackSetting.available:
            _kwargs["slack"] = True
        if TelegramSetting.available:
            _kwargs["telegram"] = True
        if DiscordSetting.available:
            _kwargs["discord"] = True
        if KookSetting.available:
            _kwargs["kook"] = True
        return cls(**_kwargs)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/8 下午1:47
# @Author  : sudoskys
# @File    : __init__.py
# @Software: PyCharm

"""
这里的 schema 是为服务商的sdk照应而写的。
包含请求体，响应体，以及对齐的数据结构。
我们自己的数据结构在 sdk/schema.py 中定义。
请求时，我们需要将自己的数据结构转换为这里的数据结构。
"""
from .tee import Driver
from .tokenizer import get_tokenizer



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午7:08
# @Author  : sudoskys
# @File    : function.py
# @Software: PyCharm

__receiver__ = "llm_task"

import json
import os

from aio_pika.abc import AbstractIncomingMessage
from loguru import logger

from llmkira.middleware.chain_box import Chain, AuthReloader, ChainReloader
from llmkira.middleware.env_virtual import EnvManager
from llmkira.schema import RawMessage
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.sdk.schema import TaskBatch
from llmkira.task import Task, TaskHeader


class ChainFunc(object):
    @staticmethod
    async def reply_user(
        *, platform: str, task: TaskHeader, text: str, receiver: TaskHeader.Location
    ):
        """
        包装一下发送消息
        :param platform: Default should be `task.receiver.platform`
        :param task: 任务 header
        :param text: 文本 str
        :param receiver: 接收者 TaskHeader.Location
        """
        # 通知用户
        return await Task.create_and_send(
            queue_name=platform,
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=task.task_meta.reply_direct(chain_name=__receiver__),
                message=[
                    RawMessage(
                        user_id=task.receiver.user_id,
                        chat_id=task.receiver.chat_id,
                        text=text,
                    )
                ],
            ),
        )

    async def auth_chain(self, *, task: TaskHeader, task_batch: TaskBatch):
        """
        认证链重发注册
        """
        _task_forward: TaskHeader = task.model_copy()
        meta: TaskHeader.Meta = _task_forward.task_meta.chain(
            name=__receiver__,
            write_back=False,  # 因为是发送给自己，所以不需要写回
            release_chain=True,  # 要求释放链
        )
        """添加认证链并重置路由数据"""
        if meta.run_step_limit < meta.run_step_already:
            return logger.debug("Reject Invalid Request, Already Reach Limit")
        """拒绝不合法的请求"""
        meta.verify_uuid = meta.get_verify_uuid(task_batch=task_batch)
        # 注册本地部署点
        task_id = await AuthReloader(uid=_task_forward.receiver.uid).save_auth(
            chain=Chain.create(
                thead_uuid=_task_forward.task_meta.task_uuid,
                uuid=meta.verify_uuid,
                creator_uid=_task_forward.receiver.uid,
                channel=__receiver__,
                # 重要：转发回来这里
                arg=TaskHeader(
                    sender=_task_forward.sender,
                    receiver=_task_forward.receiver,
                    task_meta=meta,
                    message=[],
                ),
                expire=60 * 60 * 2,
            )
        )
        await self.reply_user(
            platform=_task_forward.receiver.platform,
            receiver=_task_forward.receiver,
            task=task,
            text=f"🔑 Type `/auth {task_id}` to run `{task_batch.get_batch_name()}`"
            f"\ntry `!auth {task_id}` when no slash command",
        )
        return logger.trace("Auth Chain Resign Success")

    async def resign_chain(
        self, task: TaskHeader, parent_func: str, repeatable: bool, deploy_child: int
    ):
        """
        子链孩子函数，请注意，此处为高风险区域，预定一下函数部署点位
        :param task: 任务
        :param parent_func: 父函数
        :param repeatable: 是否可重复
        :param deploy_child: 是否部署子链
        """
        _task_forward: TaskHeader = task.model_copy()
        # 添加认证链并重置路由数据
        meta: TaskHeader.Meta = _task_forward.task_meta.chain(
            name=__receiver__, write_back=True, release_chain=True
        )
        if meta.run_step_limit < meta.run_step_already:
            return logger.debug("Reject Invalid Request, Already Reach Limit")
        if deploy_child == 0:
            return logger.debug("Parent Function Reject Resign Its Child Chain")
        """拒绝不合法的请求"""
        _task_forward.task_meta = meta
        try:
            if not repeatable:
                _task_forward.task_meta.function_list = [
                    item
                    for item in _task_forward.task_meta.function_list
                    if item.name != parent_func
                ]
                logger.trace("Remove Used Function From Function List")
        except Exception as e:
            logger.error(e)
            logger.warning(f"[362211]Remove function {parent_func} failed")
        """拒绝不合法的请求"""
        # 注册本地部署点
        _task = TaskHeader(
            sender=_task_forward.sender,
            receiver=_task_forward.receiver,
            task_meta=meta,
            message=[],
        )
        await ChainReloader(uid=_task_forward.receiver.uid).add_task(
            chain=Chain.create(
                thead_uuid=_task_forward.task_meta.task_uuid,
                creator_uid=_task_forward.receiver.uid,
                channel=_task_forward.receiver.platform,
                expire=60 * 60 * 2,
                arg=_task,
            )
        )
        logger.debug(f"Resign Chain Success --from_function {parent_func}")
        return None


class FunctionReceiver(object):
    """
    receive message from any platform
    """

    def __init__(self):
        self.task = Task(queue=__receiver__)

    @staticmethod
    async def run_pending_task(task: TaskHeader, pending_task: TaskBatch):
        """
        如果执行异常，必须抛出异常，否则会导致任务无法结束
        如果重发认证，不需要结束任务
        :param task: 任务
        :param pending_task: 待执行的函数
        :return: None
        """
        assert isinstance(pending_task, TaskBatch), "pending task type error"
        chain_func = ChainFunc()
        # Parse Function Call
        try:
            _arg = json.loads(pending_task.get_batch_args())
        except json.JSONDecodeError as decode_error:
            logger.warning("Function Arguments is not json format")
            await chain_func.reply_user(
                platform=task.receiver.platform,
                receiver=task.receiver,
                task=task,
                text=f"🔭 Sorry function `{pending_task.get_batch_name()}` "
                f"arguments is not json format"
                f"\narguments {pending_task.get_batch_args()}",
            )
            raise decode_error
        # Get Function Object
        _tool_cls = ToolRegister().get_tool(name=pending_task.get_batch_name())
        if not _tool_cls:
            logger.warning(f"Not found function {pending_task.get_batch_name()}")
            await chain_func.reply_user(
                platform=task.receiver.platform,
                receiver=task.receiver,
                task=task,
                text=f"🔭 Sorry function `{pending_task.get_batch_name()}` executor not found",
            )
            raise ModuleNotFoundError(
                f"Function {pending_task.get_batch_name()} not found"
            )
        # Run Function
        _tool_obj = _tool_cls()
        if _tool_obj.require_auth:
            if task.task_meta.verify_uuid:
                # 是携带密钥的函数，是预先构建的可信任务头
                task.task_meta.verify_uuid = None
            else:
                # 需要认证，预构建携带密钥的待发消息并回退
                await chain_func.auth_chain(task=task, task_batch=pending_task)
                return logger.info(
                    f"[Resign Auth] \n--auth-require {pending_task.get_batch_name()} require."
                )
        # Get Env
        _env_dict = await EnvManager.from_uid(uid=task.receiver.uid).get_env_list(
            name_list=_tool_obj.env_list
        )
        assert isinstance(_env_dict, dict), "unexpected env dict? it should be dict..."
        # Resign Chain
        if (
            task.task_meta.resign_next_step or task.task_meta.is_complete(num_end=1)
        ) and not _tool_obj.repeatable:
            logger.debug(f"Function {pending_task.get_batch_name()} need resign chain")
            await chain_func.resign_chain(
                task=task,
                parent_func=pending_task.get_batch_name(),
                repeatable=_tool_obj.repeatable,
                deploy_child=_tool_obj.deploy_child,
            )
        # 运行函数, 传递模型的信息，以及上一条的结果的openai raw信息
        run_result = await _tool_obj.load(
            task=task,
            receiver=task.receiver,
            arg=_arg,
            env=_env_dict,
            pending_task=pending_task,
            refer_llm_result=task.task_meta.llm_result,
        )
        # 更新任务状态
        await task.task_meta.complete_task(
            task_batch=pending_task, run_result=run_result
        )
        return run_result

    async def process_function_call(self, message: AbstractIncomingMessage):
        """
        定位，解析，运行函数。要求认证，或申请结束/继续指标。
        :param message: message from queue
        :return: None
        """
        # Parse Message
        if os.getenv("LLMBOT_STOP_REPLY") == "1":
            return None
        task: TaskHeader = TaskHeader.model_validate_json(
            json_data=message.body.decode("utf-8")
        )
        # Get Function Call
        pending_task = await task.task_meta.work_pending_task(
            verify_uuid=task.task_meta.verify_uuid
        )
        if not pending_task:
            logger.trace("No Function Call")
            return None
        pending_task: TaskBatch
        logger.debug("Received A Batch FunctionRequest")
        try:
            await self.run_pending_task(task=task, pending_task=pending_task)
        except Exception as e:
            await task.task_meta.complete_task(task_batch=pending_task, run_result=e)
            logger.error(f"Function Call Error {e}")
            raise e
        finally:
            logger.trace("Function Call Finished")

    async def on_message(self, message: AbstractIncomingMessage):
        """
        处理message
        :param message: message from queue
        :return: None
        """
        try:
            await self.process_function_call(message=message)
        except Exception as e:
            logger.exception(f"Function Receiver Error {e}")
            await message.reject(requeue=False)
            raise e
        else:
            await message.ack(multiple=False)

    async def function(self):
        logger.success("Receiver Runtime:Function Fork Cpu start")
        await self.task.consuming_task(self.on_message)



# -*- coding: utf-8 -*-
# @Time    : 2023/10/21 下午8:02
# @Author  : sudoskys
# @File    : creat_message.py
# @Software: PyCharm
from enum import Enum
from typing import Union

import emoji


class SlackEmoji(Enum):
    robot = ":robot_face:"
    check = ":white_check_mark:"
    error = ":x:"
    pin = ":pushpin:"
    thumbs_up = ":thumbsup:"
    thumbs_down = ":thumbsdown:"
    eyes = ":eyes:"
    gear = ":gear:"
    pencil = ":pencil:"
    moai = ":moai:"
    telescope = ":telescope:"
    hammer = ":hammer:"
    warning = ":warning:"


class ChatMessageCreator(object):
    WELCOME_BLOCK = {
        "type": "section",
        "text": {
            "type": "mrkdwn",
            "text": (
                "Welcome to Slack! :wave: We're so glad you're here. :blush:\n\n"
                "*Get started by completing the steps below:*"
            ),
        },
    }
    DIVIDER_BLOCK = {"type": "divider"}

    def __init__(self,
                 channel,
                 user_name: str = None,
                 thread_ts: str = None,

                 ):
        self.channel = channel
        self.username = user_name if user_name else "OAIbot"
        self.icon_emoji = ":robot_face:"
        self.timestamp = ""
        self.thread_ts = thread_ts
        self.blocks = []

    @staticmethod
    def build_block(text, msg_type: str = "section"):
        """
        create a section block
        """
        _block = {
            "type": msg_type,
            "text": {
                "type": "mrkdwn",
                "text": text,
            },
        }
        return _block

    def update_content(self, message_text):
        """
        Message text in markdown format
        """
        self.blocks.append(self.build_block(message_text))
        return self

    def update_emoji(self, emoji_name: Union[str, SlackEmoji]):
        if isinstance(emoji_name, SlackEmoji):
            self.icon_emoji = emoji_name.value
            return self
        emoji_name = emoji_name.strip()
        if ":" in emoji_name:
            self.icon_emoji = emoji_name
        else:
            _emoji = emoji.demojize(emoji_name)
            if _emoji.endswith(":"):
                self.icon_emoji = _emoji
        return self

    def get_message_payload(self, message_text=None):
        if not self.blocks:
            raise ValueError("Message cannot be empty")
        _arg = {
            "ts": self.timestamp,
            "channel": self.channel,
            "username": self.username,
            "icon_emoji": self.icon_emoji,
            "blocks": self.blocks,
        }
        if self.thread_ts:
            _arg["thread_ts"] = self.thread_ts
        if message_text:
            _arg["text"] = message_text
        return _arg



# -*- coding: utf-8 -*-
# @Time    : 2023/10/26 下午11:32
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm


# 这里是分发授权的地方。

__provider__ = {}

import os
from functools import wraps
from typing import Type

from dotenv import load_dotenv
from loguru import logger

from .schema import BaseProvider, ProviderSettingObj

load_dotenv()

pkg_path = os.path.dirname(__file__)
pkg_name = os.path.basename(pkg_path)

import os

__all__ = [
    f.strip(".py")
    for f in os.listdir(os.path.abspath(os.path.dirname(__file__)))
    if f.endswith('.py') and "_" not in f
]


def resign_provider():
    """
    装饰器
    """

    def decorator(func: Type[BaseProvider]):
        if issubclass(func, BaseProvider):
            logger.success(f"📦 [Plugin Provider Add] {func}")
            __provider__[func.name.__str__().upper()] = func

        @wraps(func)
        def wrapper(*args, **kwargs):
            # 调用执行函数，中间人
            return func(*args, **kwargs)

        return wrapper

    return decorator


from .public import PublicProvider
from .private import PrivateProvider


def setup():
    _provider_name = ProviderSettingObj.provider
    if __provider__.get(_provider_name, None):
        logger.success(f"\n️🥕 ️Use Service Provider Named:{_provider_name}")
        return __provider__.get(_provider_name)
    else:
        logger.warning(f"\n⚠ No Select Service Provider Named:{_provider_name},Use Public Provider")
        return PublicProvider


loaded_provider: Type[BaseProvider] = setup()



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:31
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
import time
from typing import TYPE_CHECKING, Literal, Type, Optional
from typing import Union, List

import nest_asyncio
from pydantic import field_validator, ConfigDict, Field, BaseModel

from .sdk.endpoint.tokenizer import get_tokenizer
from .sdk.schema import File, generate_uid, UserMessage, Message

if TYPE_CHECKING:
    from .task import TaskHeader

nest_asyncio.apply()


class RawMessage(BaseModel):
    user_id: Union[int, str] = Field(None, description="user id")
    chat_id: Union[int, str] = Field(None, description="guild id(channel in dm)/Telegram chat id")
    thread_id: Optional[Union[int, str]] = Field(None, description="channel id/Telegram thread")

    text: str = Field("", description="文本")
    file: List["File"] = Field([], description="文件")

    created_at: str = Field(default=str(int(time.time())), description="创建时间")
    only_send_file: bool = Field(default=False, description="Send file only")

    sign_loop_end: bool = Field(default=False, description="要求其他链条不处理此消息，用于拦截器开发")
    sign_fold_docs: bool = Field(default=False, description="是否获取元数据")
    extra_kwargs: dict = Field(default={}, description="extra kwargs for loop")
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    @field_validator("text")
    def check_text(cls, v):
        if v == "":
            v = ""
        if len(v) > 4096:
            v = v[:4090]
        return v

    def format_user_message(
            self,
            role: Literal["system", "assistant", "user", "function", "tool"] = "user",
            name: str = None
    ) -> UserMessage:
        """
        OpenAI 格式化
        """
        return UserMessage(
            role=role,
            name=name,
            content=self.text,
        ).update_meta(meta=Message.Meta(
            retrieval=self.sign_fold_docs,
            files=self.file,
            index_id=self.thread_id or generate_uid(),
        )
        )

    @classmethod
    def format_openai_message(
            cls,
            message: Type[Message],
            locate: "TaskHeader.Location"
    ) -> "RawMessage":
        """
        用于 OpenAI 消息转换回复给用户
        """
        # IMPORT FROM llmkira.task
        return cls(
            user_id=locate.user_id,
            text=message.content,
            chat_id=locate.chat_id,
            created_at=str(int(time.time()))
        )


def singleton(cls):
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner


class Scraper(BaseModel):
    """
    刮削器
    始终按照顺序排列，削除得分低的条目
    Scraper is a class that sorts a list of messages_box by their score.
    """

    class Sorter(BaseModel):
        message: Union[Type[Message], Message]
        score: float
        order: int

        def get_message(self) -> Union[Type[Message], Message]:
            return self.message

    messages_box: List[Sorter] = []
    # 最大消息数
    max_messages: int = 12
    # 计数器
    tick: int = 0

    # 方法：添加消息
    def add_message(self, message: Union[Type[Message], Message], score: float):
        if not hasattr(message, "content"):
            """message has no content"""
            return None
        self.messages_box.append(self.Sorter(message=message, score=score, order=self.tick))
        self.tick += 1
        # 按照顺序排序
        self.messages_box.sort(key=lambda x: x.order)
        while len(self.messages_box) > self.max_messages:
            self.messages_box.pop(0)

    # 方法：获取消息
    def get_messages(
            self
    ) -> List[Union[Type[Message], Message]]:
        _message: List[Union[Type[Message], Message]] = []
        for sorter in self.messages_box:
            assert isinstance(sorter.message, Message), f"message type error {type(sorter.message)}"
            _message.append(sorter.message)
        return _message

    def build_messages(self):
        # 只取三个，末位匹配
        _message = self.get_messages()
        if len(_message) < 3:
            return _message
        _build = []
        _must = _message[-3:]
        _check_list = _message[:-3]
        _match_sentence = _message[-1:][0].content
        for item_obj in _check_list:
            _build.append(item_obj)
            # NOTICE 此处被内存优化，不再使用
            """
            if Sim.cos ion_similarity(pre=_match_sentence, aft=item_obj.content) < 0.9:
                _build.append(item_obj)
            else:
                pass
                # logger.warning(f"ignore sim item {item_obj}")
            """
        _build.extend(_must)
        return _build

    # 方法：获取消息数
    def get_num_messages(self) -> int:
        return len(self.messages_box)

    def fold_message(self):
        """
        以折叠形态出击
        """
        for sorter in self.messages_box:
            if sorter.get_message().get_meta().retrieval:
                sorter.message = sorter.message.fold
        return self

    # 方法：清除消息到负载
    def reduce_messages(self,
                        *,
                        limit: int = 2048,
                        model_name: str
                        ):
        """
        对齐 0.7 倍极限的消息记录
        :param limit: 最大字数
        :param model_name: 模型名称
        :return: None
        """
        # 预留位
        if limit > 1000:
            limit = limit * 0.8
        else:
            limit = limit * 0.7
        # 执行删除操作
        if get_tokenizer(model_name=model_name).num_tokens_from_messages(
                messages=self.get_messages(),
                model=model_name
        ) > limit:
            # 从最旧开始删除
            self.messages_box.sort(key=lambda x: x.order)
            while get_tokenizer(model_name=model_name).num_tokens_from_messages(
                    messages=self.get_messages(),
                    model=model_name
            ) > limit:
                if len(self.messages_box) > 1:
                    self.messages_box.pop(0)
                else:
                    self.messages_box[0].message.content = self.messages_box[0].message.content[:limit]
        # 按照顺序排序
        self.messages_box.sort(key=lambda x: x.order)



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:43
# @Author  : sudoskys
# @File    : redis.py
# @Software: PyCharm
import json
from typing import Optional, Tuple, List, Union

import redis
from loguru import logger
from redis.asyncio.client import Redis
from redis.asyncio.connection import ConnectionPool

from .base import AbstractDataClass, PREFIX


class RedisClientWrapper(AbstractDataClass):
    """
    Redis 数据类
    """

    def __init__(self, backend, prefix=PREFIX):
        self.prefix = prefix
        self.connection_pool = redis.asyncio.ConnectionPool.from_url(backend)
        self._redis = redis.asyncio.Redis(connection_pool=self.connection_pool)

    async def ping(self):
        return await self._redis.ping()

    def update_backend(self, backend):
        self.connection_pool = ConnectionPool.from_url(backend)
        self._redis = Redis(connection_pool=self.connection_pool)
        return True

    async def set_data(self, key, value, timeout=None):
        if isinstance(value, (dict, list)):
            value = json.dumps(value)
        return await self._redis.set(
            name=f"{self.prefix}{key}", value=value, ex=timeout
        )

    async def read_data(self, key) -> Optional[Union[str, dict, int]]:
        data = await self._redis.get(self.prefix + str(key))
        if data is not None:
            try:
                data = json.loads(data)
            except Exception as ex:
                logger.trace(ex)
                pass
        return data

    async def lpush_data(self, key, value):
        """
        从左侧插入数据
        :param key: str
        :param value: json
        """
        # 验证是否可以被json序列化
        if isinstance(value, (dict, list)):
            value = json.dumps(value)
        return await self._redis.lpush(self.prefix + str(key), value)

    async def lpop_data(self, key) -> Optional[dict]:
        _data = await self._redis.lpop(self.prefix + str(key))
        if _data:
            _data = _data.decode("utf-8")
            try:
                _data = json.loads(_data)
            except json.JSONDecodeError:
                pass
            return _data
        return None

    async def lrange_data(self, key, start_end: Tuple[int, int] = (0, -1)) -> List[str]:
        _items = await self._redis.lrange(
            self.prefix + str(key), start=start_end[0], end=start_end[1]
        )
        items = [m.decode("utf-8") for m in _items[::-1]]
        return items



# -*- coding: utf-8 -*-
# @Time    : 2023/8/15 下午10:34
# @Author  : sudoskys
# @File    : base.py
# @Software: PyCharm
from pydantic import field_validator, ConfigDict, model_validator

__version__ = "0.0.1"

from typing import Union, List, Optional, Literal, Type

import httpx
import pydantic
from dotenv import load_dotenv
from loguru import logger
from pydantic import BaseModel, Field, PrivateAttr

from ..schema import LlmResult, LlmRequest
from ..tee import Driver
from ..tokenizer import get_tokenizer
from ...adapter import SCHEMA_GROUP, SingleModel
from ...error import ValidationError
from ...network import request
from ...schema import Message, Function, ToolChoice, AssistantMessage, Tool, BaseFunction, ToolMessage

load_dotenv()


class OpenaiResult(LlmResult):
    class Usage(BaseModel):
        prompt_tokens: int
        completion_tokens: int
        total_tokens: int

    class Choices(BaseModel):
        index: int
        message: AssistantMessage
        finish_reason: str
        """
        The reason the model stopped generating tokens. This will be stop if the model hit a natural stop point or 
        a provided stop sequence, length if the maximum number of tokens specified in the request was reached, 
        content_filter if content was omitted due to a flag from our content filters, tool_calls if the model called 
        a tool, or function_call (deprecated) if the model called a function.
        """
        delta: dict = None

        @property
        def sign_function(self):
            return bool(
                "function_call" == self.finish_reason
                or
                "tool_calls" == self.finish_reason
            )

    id: str
    object: str
    created: int
    model: str
    system_fingerprint: str = Field(default=None, alias="system_prompt_fingerprint")
    choices: List[Choices]
    usage: Usage
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    @property
    def result_type(self):
        return self.object

    @property
    def default_message(self):
        return self.choices[0].message


class Openai(LlmRequest):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    _config: Driver = PrivateAttr(default=None)
    """模型信息和配置"""
    messages: List[Union["Message", Type["Message"]]]
    temperature: Optional[float] = 1
    n: Optional[int] = 1
    top_p: Optional[float] = None
    stop: Optional[Union[str, List[str]]] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    seed: Optional[int] = None
    """基础设置"""
    stream: Optional[bool] = False
    """暂时不打算用的流式"""
    logit_bias: Optional[dict] = None
    """暂时不打算用的logit_bias"""
    user: Optional[str] = None
    """追踪 User"""
    response_format: Optional[dict] = None
    """回复指定的格式，See: https://platform.openai.com/docs/api-reference/chat/create"""
    # 函数
    functions: Optional[List["Function"]] = None
    """deprecated"""
    function_call: Optional[Union["BaseFunction", Literal["auto", "none"]]] = None
    """deprecated"""
    # 工具
    tools: Optional[List["Tool"]] = None
    tool_choice: Optional[Union["ToolChoice", Literal["auto", "none"]]] = None
    """工具调用"""

    # 注意，我们
    """
    如果是 vison ，需要转换消息。
    注意验证 tool choice 的字段。
    """

    @model_validator(mode="before")
    def fix_tool(cls, values):
        if not values.get("tools"):
            values["tools"] = None
        else:
            assert isinstance(values.get("tools"), list)

        if not values.get("functions"):
            values["functions"] = None
        else:
            assert isinstance(values.get("functions"), list)
        if values.get("tools") and values.get("functions"):
            logger.warning("sdk param validator:'functions' and 'tools' cannot both be provided. ignoring 'functions'")
            values["functions"] = None
            values["function_call"] = None
        """Deprecated by openai"""
        return values

    @property
    def schema_map(self) -> dict:
        return {
            "model": True,
            "messages": True,
            "temperature": True,
            "top_p": True,
            "n": True,
            "stop": True,
            "max_tokens": True,
            "seed": True,
            "presence_penalty": True,
            "frequency_penalty": True,
            "stream": True,
            "logit_bias": True,
            "user": True,
            "response_format": True,
            "functions": True,
            "function_call": True,
            "tools": True,
            "tool_choice": True,
        }

    @staticmethod
    def sort_insert_message(message_list: List[Message]):
        """
        排序插入
        :param message_list:
        :return:
        """
        origin_message_list = message_list.copy()
        ordered_message_list = []
        child_resp = {}
        for message in message_list:
            if isinstance(message, ToolMessage):
                if message.tool_call_id in child_resp:
                    child_resp[message.tool_call_id].append(message)
                else:
                    child_resp[message.tool_call_id] = [message]
                message_list.remove(message)
        # 在不存在 child 的情况下，修补
        for message in message_list:
            # 存放一个
            ordered_message_list.append(message)
            if isinstance(message, AssistantMessage) and message.tool_calls:
                # 需要修补的消息
                for tool in message.tool_calls:
                    if tool.id in child_resp:
                        # 匹配到了 对应的 child
                        ordered_message_list.extend(child_resp[tool.id])
                    else:
                        # 没有匹配到对应的 child
                        ordered_message_list.append(
                            ToolMessage(
                                tool_call_id=tool.id,
                                content="[On Queue]"
                            )
                        )
        # 修补完毕
        if len(origin_message_list) != len(ordered_message_list):
            logger.warning(f"message_list is not match, origin:{origin_message_list}, orderd:{ordered_message_list}")
        return ordered_message_list

    def create_params(self):
        # 修补消息
        self.messages = self.sort_insert_message(message_list=self.messages)
        # 获取已经传递的工具模型
        _arg = self.model_dump(
            exclude_none=True,
            include=self.schema_map
        )
        assert "messages" in _arg, "messages is required"
        _arg["model"] = self.model
        _arg = {
            k: v
            for k, v in _arg.items()
            if v is not None
        }
        return _arg

    @field_validator("presence_penalty")
    def check_presence_penalty(cls, v):
        if not (2 > v > -2):
            raise ValidationError("presence_penalty must be between -2 and 2")
        return v

    @field_validator("stop")
    def check_stop(cls, v):
        if isinstance(v, list) and len(v) > 4:
            raise ValidationError("stop list length must be less than 4")
        return v

    @field_validator("temperature")
    def check_temperature(cls, v):
        if not (2 > v > 0):
            raise ValidationError("temperature must be between 0 and 2")
        return v

    async def create(self,
                     **kwargs
                     ) -> OpenaiResult:
        """
        请求
        :return:
        """
        """
        Docs:https://platform.openai.com/docs/api-reference/chat/create
        """
        # Check Token Limit
        num_tokens_from_messages = get_tokenizer(model_name=self.model).num_tokens_from_messages(
            messages=self.messages,
            model=self.model,
        )
        if num_tokens_from_messages > SCHEMA_GROUP.get_token_limit(model_name=self.model):
            raise ValidationError("messages_box num_tokens > max_tokens")
        # 返回请求
        headers = {
            "User-Agent": "Mozilla/5.0",
            "Authorization": f"Bearer {self._config.api_key}",
            "api-key": f"{self._config.api_key}",
        }
        if self._config.org_id:
            headers["Openai-Organization"] = self._config.org_id
        try:
            logger.debug(f"[Openai request] {self.create_params()}")
            _response = await request(
                method="POST",
                url=self._config.endpoint,
                data=self.create_params(),
                headers=headers,
                proxy=self.proxy_address(),
                json_body=True
            )
            assert _response, ValidationError("response is empty")
            logger.debug(f"[Openai response] {_response}")
            return_result = OpenaiResult.model_validate(_response).ack()
        except httpx.ConnectError as e:
            logger.error(f"[Openai connect error] {e}")
            raise e
        except pydantic.ValidationError as e:
            logger.error(f"[Api format error] {e}")
            raise e
        if self.echo:
            logger.info(f"[Openai Raw response] {return_result}")
        return return_result


SCHEMA_GROUP.add_model(
    models=[
        SingleModel(
            llm_model="chatglm3",
            token_limit=4096,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="function_call",
            exception=None
        ),
        SingleModel(
            llm_model="chatglm3-16k",
            token_limit=16384,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="function_call",
            exception=None
        ),
    ]
)

SCHEMA_GROUP.add_model(
    models=[
        SingleModel(
            llm_model="gpt-3.5-turbo-1106",
            token_limit=16384,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-3.5-turbo",
            token_limit=4096,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-3.5-turbo-16k",
            token_limit=16384,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-3.5-turbo-0613",
            token_limit=4096,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-3.5-turbo-16k-0613",
            token_limit=16384,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-4",
            token_limit=8192,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-4-32k",
            token_limit=32768,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-4-0613",
            token_limit=8192,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-4-vision-preview",
            token_limit=128000,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        ),
        SingleModel(
            llm_model="gpt-4-1106-preview",
            token_limit=128000,
            request=Openai,
            response=OpenaiResult,
            schema_type="openai",
            func_executor="tool_call",
            exception=None
        )
    ]
)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/8 下午8:05
# @Author  : sudoskys
# @File    : test_file.py
# @Software: PyCharm
import sys

sys.path.append("..")

from llmkira.sdk.schema import File


def test_schema():
    File(
        file_id=None, file_name="test"
    )



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 下午10:46
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import atexit
from typing import List

import hikari
from hikari.impl import ProxySettings
from loguru import logger

from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.receiver import function
from llmkira.receiver.receiver_client import BaseReceiver, BaseSender
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.schema import File, Message
from llmkira.sdk.utils import sync
from llmkira.setting.discord import BotSetting
from llmkira.task import Task, TaskHeader

__receiver__ = "discord_hikari"

from llmkira.middleware.router.schema import router_set

router_set(role="receiver", name=__receiver__)

discord_rest: hikari.RESTApp = hikari.RESTApp(
    proxy_settings=ProxySettings(
        url=BotSetting.proxy_address
    )
)


class DiscordSender(BaseSender):
    """
    平台路由
    """

    def __init__(self):
        self.bot = None

    def acquire(self):
        self.bot: hikari.impl.RESTClientImpl = discord_rest.acquire(
            token=BotSetting.token,
            token_type=hikari.TokenType.BOT
        )

    async def file_forward(self, receiver: TaskHeader.Location, file_list: List[File]):
        for file_obj in file_list:
            # URL FIRST
            if file_obj.file_url:
                async with self.bot as client:
                    client: hikari.impl.RESTClientImpl
                    _reply = None
                    if receiver.thread_id != receiver.chat_id:
                        _reply = await client.fetch_message(
                            channel=receiver.thread_id,
                            message=receiver.message_id
                        )
                    await client.create_message(
                        channel=receiver.thread_id,
                        attachment=hikari.files.URL(file_obj.file_url, filename=file_obj.file_name),
                        reply=_reply,
                    )
                break
            # DATA
            _data: File.Data = sync(file_obj.raw_file())
            if not _data:
                logger.error(f"file download failed {file_obj.file_id}")
                continue
            file_warp = hikari.files.Bytes(_data.file_data, file_obj.file_name, mimetype="image/png")
            if file_obj.file_name.endswith((".jpg", ".png")):
                async with self.bot as client:
                    client: hikari.impl.RESTClientImpl
                    _reply = None
                    if receiver.thread_id != receiver.chat_id:
                        _reply = await client.fetch_message(
                            channel=receiver.thread_id,
                            message=receiver.message_id
                        )
                    await client.create_message(
                        channel=receiver.thread_id,
                        embed=hikari.EmbedImage(
                            resource=file_warp,
                        ),
                        reply=_reply
                    )
            else:
                async with self.bot as client:
                    client: hikari.impl.RESTClientImpl
                    _reply = None
                    if receiver.thread_id != receiver.chat_id:
                        _reply = await client.fetch_message(
                            channel=receiver.thread_id,
                            message=receiver.message_id
                        )
                    await client.create_message(
                        channel=receiver.thread_id,
                        attachment=file_warp,
                        reply=_reply
                    )

    async def forward(self, receiver: TaskHeader.Location, message: List[RawMessage]):
        """
        插件专用转发，是Task通用类型
        """
        for item in message:
            await self.file_forward(
                receiver=receiver,
                file_list=item.file
            )
            if item.only_send_file:
                continue
            async with self.bot as client:
                client: hikari.impl.RESTClientImpl
                _reply = None
                if receiver.thread_id != receiver.chat_id:
                    _reply = await client.fetch_message(
                        channel=receiver.thread_id,
                        message=receiver.message_id
                    )
                await client.create_message(
                    channel=receiver.thread_id,
                    content=item.text,
                    reply=_reply
                )

    async def reply(self, receiver: TaskHeader.Location, message: List[Message], reply_to_message: bool = True):
        """
        模型直转发，Message是Openai的类型
        """
        for item in message:
            raw_message = await self.loop_turn_from_openai(platform_name=__receiver__, message=item, locate=receiver)

            await self.file_forward(
                receiver=receiver,
                file_list=raw_message.file
            )
            if raw_message.only_send_file:
                continue
            if not raw_message.text:
                continue
            assert raw_message.text, f"message content is empty"
            async with self.bot as client:
                client: hikari.impl.RESTClientImpl
                _reply = None
                if receiver.thread_id != receiver.chat_id:
                    _reply = await client.fetch_message(
                        channel=receiver.thread_id,
                        message=receiver.message_id
                    )
                await client.create_message(
                    channel=receiver.thread_id,
                    content=raw_message.text,
                    reply=_reply
                )
        return logger.trace(f"reply message")

    async def error(self, receiver: TaskHeader.Location, text):
        async with self.bot as client:
            client: hikari.impl.RESTClientImpl
            _reply = None
            if receiver.thread_id != receiver.chat_id:
                _reply = await client.fetch_message(
                    channel=receiver.thread_id,
                    message=receiver.message_id
                )
            await client.create_message(
                channel=receiver.thread_id,
                content=text,
                reply=_reply
            )

    async def function(self,
                       receiver: TaskHeader.Location,
                       task: TaskHeader,
                       llm: OpenaiMiddleware,
                       llm_result: LlmResult
                       ):
        task_batch = llm_result.default_message.get_executor_batch()
        verify_map = await self.push_task_create_message(
            task=task,
            task_batch=task_batch,
            llm_result=llm_result,
            receiver=receiver
        )
        new_receiver = task.receiver.model_copy()
        new_receiver.platform = __receiver__
        """更新接收者为当前平台，便于创建的函数消息能返回到正确的客户端"""
        new_meta = task.task_meta.pack_loop(
            plan_chain_pending=task_batch,
            verify_map=verify_map
        )
        """克隆元数据为当前平台"""

        await Task(queue=function.__receiver__).send_task(
            task=TaskHeader.from_function(
                llm_result=llm_result,
                task_meta=new_meta,
                receiver=new_receiver,
                message=task.message
            )
        )
        """发送打包后的任务数据"""
        del new_meta
        del task
        """清理内存"""


__sender__ = DiscordSender()


class DiscordReceiver(BaseReceiver):
    """
    receive message from telegram
    """

    async def discord(self):
        self.set_core(sender=__sender__, task=Task(queue=__receiver__))
        if not BotSetting.available:
            logger.warning("Receiver Runtime:Discord Setting empty")
            return None
        await discord_rest.start()
        __sender__.acquire()
        try:
            logger.success("Receiver Runtime:Discord start")
            await self.task.consuming_task(self.on_message)
        except KeyboardInterrupt:
            logger.warning("Discord Receiver shutdown")
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            await discord_rest.close()


if BotSetting.available:
    @atexit.register
    def __clean():
        try:
            sync(discord_rest.close())
        except Exception as e:
            logger.warning(f"Discord Receiver cleaning failed when exiting \n{e}")



# -*- coding: utf-8 -*-
from pydantic import field_validator, ConfigDict

__package__name__ = "llmkira.extra.plugins.sticker"
__plugin_name__ = "convert_to_sticker"
__openapi_version__ = "20231111"

import re

from llmkira.sdk import resign_plugin_executor
from llmkira.sdk.func_calling import verify_openapi_version
from llmkira.sdk.schema import File, Function

verify_openapi_version(__package__name__, __openapi_version__)
from io import BytesIO
from math import floor

from PIL import Image
from loguru import logger
from pydantic import BaseModel

from llmkira.schema import RawMessage
from llmkira.sdk.func_calling import BaseTool
from llmkira.sdk.func_calling.schema import FuncPair, PluginMetadata
from llmkira.task import Task, TaskHeader
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from llmkira.sdk.schema import TaskBatch
sticker = Function(name=__plugin_name__, description="Help user convert pictures to stickers")
sticker.add_property(
    property_name="yes_no",
    property_description="Is this run allowed (yes/no) If there a picture, please say yes",
    property_type="string",
    required=True
)
sticker.add_property(
    property_name="comment",
    property_description="thanks for this run",
    property_type="string",
    required=True
)


class Sticker(BaseModel):
    yes_no: str
    comment: str = "done"
    model_config = ConfigDict(extra="allow")

    @field_validator("yes_no")
    def delay_validator(cls, v):
        if v != "yes":
            v = "no"
        return v


@resign_plugin_executor(function=sticker)
async def resize_image(photo):
    logger.debug(f"Plugin --resize_image")
    image = Image.open(photo)
    if (image.width and image.height) < 512:
        size1 = image.width
        size2 = image.height
        if image.width > image.height:
            scale = 512 / size1
            size1new = 512
            size2new = size2 * scale
        else:
            scale = 512 / size2
            size1new = size1 * scale
            size2new = 512
        size1new = floor(size1new)
        size2new = floor(size2new)
        size_new = (size1new, size2new)
        image = image.resize(size_new)
    else:
        maxsize = (512, 512)
        image.thumbnail(maxsize)

    return image


class StickerTool(BaseTool):
    """
    搜索工具
    """
    function: Function = sticker
    keywords: list = ["转换", "贴纸", ".jpg", "图像", '图片']
    file_match_required: Optional[re.Pattern] = re.compile(r"(.+).jpg|(.+).png|(.+).jpeg|(.+).gif|(.+).webp|(.+).svg")

    def pre_check(self):
        return True

    def func_message(self, message_text, **kwargs):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self,
                     task: "TaskHeader", receiver: "TaskHeader.Location",
                     exception,
                     env: dict,
                     arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                     **kwargs
                     ):
        _meta = task.task_meta.reply_notify(
            plugin_name=__plugin_name__,
            callback=[TaskHeader.Meta.Callback.create(
                tool_call_id=pending_task.get_batch_id(),
                function_response=f"Run Failed {exception}",
                name=__plugin_name__
            )],
            write_back=False,
            release_chain=True
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=f"🍖{__plugin_name__} Run Failed：{exception}"
                    )
                ]
            )
        )

    async def callback(self,
                       task: "TaskHeader", receiver: "TaskHeader.Location",
                       env: dict,
                       arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                       **kwargs
                       ):
        return None

    async def run(self,
                  task: "TaskHeader", receiver: "TaskHeader.Location",
                  arg: dict, env: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                  ):
        """
        处理message，返回message
        """
        _file = []
        for item in task.message:
            assert isinstance(item, RawMessage), "item must be RawMessage"
            if item.file:
                for i in item.file:
                    _file.append(i)
        _set = Sticker.model_validate(arg)
        _file_obj = [await i.raw_file() for i in sorted(set(_file), key=_file.index)]
        # 去掉None
        _file_obj = [item for item in _file_obj if item]
        _result = []
        if not _file_obj:
            return
        for item in _file_obj:
            image = await resize_image(BytesIO(item.file_data))
            file = BytesIO()
            file.name = "sticker.webp"
            image.save(file, "WEBP")
            file.seek(0)
            file_obj = await File.upload_file(file_name="sticker.webp",
                                              file_data=file.getvalue(),
                                              creator_uid=receiver.uid
                                              )
            _result.append(file_obj)
        # META
        _meta = task.task_meta.reply_message(
            plugin_name=__plugin_name__,
            callback=[
                TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response="Run Success",
                    tool_call_id=pending_task.get_batch_id()
                )
            ],
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        file=_result,
                        text="Here is your sticker!"
                    )
                ]
            )
        )

        logger.debug("Plugin --convert_to_sticker {}".format(_set.yes_no))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Convert pictures to stickers",
    usage="需要回复图片，用中文指示转换贴纸即可",
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=sticker, tool=StickerTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/15 下午10:29
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm

from .func_calling import load_from_entrypoint, get_entrypoint_plugins
from .openapi.fuse import resign_plugin_executor, recover_error_plugin, get_error_plugin

"""
from .openapi.transducer import resign_transfer
from .openapi.trigger import resign_trigger
from .openapi.fuse import resign_plugin_executor, recover_error_plugin, get_error_plugin
"""



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:42
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import os
from typing import Optional

from dotenv import load_dotenv
from loguru import logger

from .base import singleton, BaseRuntime
from .mongo import MongoClientWrapper
from .redis import RedisClientWrapper
from ..utils import sync


@singleton
class MongodbRuntime(BaseRuntime):
    client: Optional["MongoClientWrapper"] = None
    init_already = False
    dsn = None

    @staticmethod
    def check_mongodb_dsn(dsn):
        client = MongoClientWrapper(dsn)
        assert sync(client.ping()), f"MongoDB {dsn} Disconnect"

    def check_mongodb(self):
        load_dotenv()
        self.dsn = os.getenv('MONGODB_DSN', "mongodb://admin:8a8a8a@localhost:27017/?authSource=admin")
        try:
            self.check_mongodb_dsn(self.dsn)
        except Exception as e:
            logger.error(f'\n⚠️ Mongodb DISCONNECT, pls set MONGODB_DSN in .env\n--error {e} \n--dsn {self.dsn}')
            raise e
        else:
            logger.success(f'🍩 MongoClientWrapper Loaded --dsn {self.dsn}')
            self.client = MongoClientWrapper(self.dsn)
            if "mongodb://admin:8a8a8a@" in self.dsn:
                logger.warning(
                    f"\n⚠️ You Are Using The Default MongoDB Password"
                    f"\nMake Sure You Handle The Port `27017` And Set Firewall Rules"
                )

    def init_mongodb(self,
                     verbose=False):
        if verbose:
            logger.info("Try To Connect To MongoDb")
        self.check_mongodb()
        self.init_already = True
        assert isinstance(self.client, MongoClientWrapper), f"MongoDb type error {type(self.client)}"
        return self.client

    def get_mongodb(self) -> "MongoClientWrapper":
        if not self.init_already:
            self.init_mongodb()
            assert isinstance(self.client, MongoClientWrapper), f"MongoDb error {type(self.client)}"
        else:
            assert isinstance(self.client, MongoClientWrapper), f"Inited MongoDb error {type(self.client)}"
        return self.client


@singleton
class RedisRuntime(BaseRuntime):
    client: Optional["RedisClientWrapper"] = None
    init_already = False
    dsn = None

    @staticmethod
    def check_redis_dsn(dsn):
        import redis
        r = redis.from_url(dsn)
        assert r.ping() is True

    def check_redis(self):
        load_dotenv()
        self.dsn = os.getenv('REDIS_DSN', 'redis://localhost:6379/0')
        try:
            self.check_redis_dsn(self.dsn)
        except Exception as e:
            logger.error(f'\n⚠️ Redis Disconnect, Please set `REDIS_DSN` env  --error {e} --dsn {self.dsn}')
            raise e
        else:
            logger.success(f'🍩 RedisClientWrapper Loaded --dsn {self.dsn}')
            self.client = RedisClientWrapper(self.dsn)
            if self.dsn == 'redis://localhost:6379/0':
                logger.warning("\n⚠️ You are using a non-password-protected local REDIS database.")

    def init_cache(self,
                   verbose=False) -> "RedisClientWrapper":
        if verbose:
            logger.info("Try To Connect To Cache")
        self.check_redis()
        self.init_already = True
        assert isinstance(self.client, RedisClientWrapper), f"Cache type error {type(self.client)}"
        return self.client

    def get_redis(self) -> "RedisClientWrapper":
        if not self.init_already:
            self.init_cache()
            assert isinstance(self.client, RedisClientWrapper), f"Cache type error {type(self.client)}"
        else:
            assert isinstance(self.client, RedisClientWrapper), f"Inited cache type error {type(self.client)}"
        return self.client


global_cache_runtime: RedisRuntime = RedisRuntime()
global_mongodb_runtime: MongodbRuntime = MongodbRuntime()



# -*- coding: utf-8 -*-
# @Time    : 2023/9/25 下午10:48
# @Author  : sudoskys
# @File    : receiver_client.py
# @Software: PyCharm
#####
# This file is not a top-level schematic file!
#####

import os
import ssl
from abc import ABCMeta, abstractmethod
from typing import Optional, Tuple, List

import httpx
import shortuuid
from aio_pika.abc import AbstractIncomingMessage
from llmkira.error import get_request_error_message, ReplyNeededError
from llmkira.middleware.chain_box import Chain, ChainReloader
from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.func_reorganize import FunctionReorganize
from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.middleware.service_provider.schema import ProviderException
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.error import RateLimitError, ServiceUnavailableError
from llmkira.sdk.func_calling import ToolRegister
from llmkira.sdk.openapi.transducer import LoopRunner
from llmkira.sdk.schema import AssistantMessage, TaskBatch
from llmkira.task import Task, TaskHeader
from loguru import logger
from pydantic import ValidationError as PydanticValidationError
from telebot import formatting


class BaseSender(object, metaclass=ABCMeta):
    @staticmethod
    async def loop_turn_from_openai(platform_name, message, locate):
        """
        将 Openai 消息传入 Receiver Loop 进行修饰
        此过程将忽略掉其他属性。只留下 content
        """
        loop_runner = LoopRunner()
        trans_loop = loop_runner.get_receiver_loop(platform_name=platform_name)
        _raw_message = RawMessage.format_openai_message(
            message=message,
            locate=locate
        )
        await loop_runner.exec_loop(
            pipe=trans_loop,
            pipe_arg={
                "message": _raw_message,
            }
        )
        arg: dict = loop_runner.result_pipe_arg
        if not arg.get("message"):
            logger.error("Message Loop Lose Message")
        raw_message: RawMessage = arg.get("message", _raw_message)
        assert isinstance(raw_message, RawMessage), f"message type error {type(raw_message)}"
        return raw_message

    @abstractmethod
    async def file_forward(self, receiver, file_list):
        raise NotImplementedError

    @abstractmethod
    async def forward(self, receiver, message):
        """
        插件专用转发，是Task通用类型
        """
        raise NotImplementedError

    @abstractmethod
    async def reply(self, receiver, message, reply_to_message: bool = True):
        """
        模型直转发，Message是Openai的类型
        """
        raise NotImplementedError

    @abstractmethod
    async def error(self, receiver, text):
        raise NotImplementedError

    async def push_task_create_message(self,
                                       *,
                                       receiver: TaskHeader.Location,
                                       task: TaskHeader,
                                       llm_result: LlmResult,
                                       task_batch: List[TaskBatch]
                                       ):
        auth_map = {}

        async def _action_block(_task_batch: TaskBatch) -> Tuple[List[str], bool]:
            _tool = ToolRegister().get_tool(_task_batch.get_batch_name())
            if not _tool:
                logger.warning(f"not found function {_task_batch.get_batch_name()}")
                return [
                    formatting.mbold("🍩 [Unknown]") + f" `{_task_batch.get_batch_name()}` "
                ], False
            tool = _tool()
            icon = "🌟"
            if tool.require_auth:
                icon = "🔐"
                auth_map[str(shortuuid.uuid()[0:5]).upper()] = _task_batch
                logger.trace(f"🔐 Auth Map {auth_map}")
            _func_tips = [
                formatting.mbold(f"{icon} [ActionBlock]") + f" `{_task_batch.get_batch_name()}` ",
                f"""```\n{_task_batch.get_batch_args()}\n```""" if not tool.silent else "",
            ]
            if tool.env_list:
                __secret__ = await EnvManager.from_uid(
                    uid=task.receiver.uid
                ).get_env_list(name_list=tool.env_list)
                # 查找是否有空
                _required_env = [
                    name
                    for name in tool.env_list
                    if not __secret__.get(name, None)
                ]
                _need_env_list = [
                    f"`{formatting.escape_markdown(name)}`"
                    for name in _required_env
                ]
                _need_env_str = ",".join(_need_env_list)
                _func_tips.append(formatting.mbold("🦴 Env required:") + f" {_need_env_str} ")
                help_docs = tool.env_help_docs(_required_env)
                _func_tips.append(formatting.mitalic(help_docs))
            return _func_tips, tool.silent

        task_message = [
            formatting.mbold("💫 Plan") + f" `{llm_result.id[-4:]}` ",
        ]
        total_silent = True
        assert isinstance(task_batch, list), f"task batch type error {type(task_batch)}"
        for _task_batch in task_batch:
            _message, _silent = await _action_block(_task_batch=_task_batch)
            if not _silent:
                total_silent = False
            if isinstance(_message, list):
                task_message.extend(_message)
        task_message_str = formatting.format_text(
            *task_message,
            separator="\n"
        )
        if not total_silent:
            await self.forward(receiver=receiver,
                               message=[
                                   RawMessage(
                                       text=task_message_str,
                                       only_send_file=False
                                   )
                               ]
                               )
        return auth_map

    @abstractmethod
    async def function(self,
                       *,
                       receiver,
                       task,
                       llm,
                       llm_result
                       ):
        raise NotImplementedError


class BaseReceiver(object):
    def __init__(self):
        self.sender: Optional[BaseSender] = None
        self.task: Optional[Task] = None

    def set_core(self, sender: BaseSender, task: Task):
        self.sender = sender
        self.task = task

    @staticmethod
    async def llm_request(
            *,
            llm_agent: OpenaiMiddleware,
            auto_write_back: bool = True,
            retrieve_message: bool = False,
            disable_function: bool = False
    ):
        """
        Openai请求
        :param llm_agent: Openai中间件
        :param auto_write_back: 是否将task携带的消息回写进消息池中，如果为False则丢弃task携带消息
        :param disable_function: 是否禁用函数
        :param retrieve_message: 是否检索消息
        :return: OpenaiResult
        校验包装，没有其他作用
        """
        try:
            _result = await llm_agent.request_openai(
                auto_write_back=auto_write_back,
                disable_function=disable_function,
                retrieve_mode=retrieve_message
            )
            return _result
        except ssl.SSLSyscallError as e:
            logger.error(f"[Network ssl error] {e},that maybe caused by bad proxy")
            raise Exception(e)
        except httpx.RemoteProtocolError as e:
            logger.error(f"[Network RemoteProtocolError] {e}")
            raise ReplyNeededError(message="Server disconnected without sending a response.")
        except ServiceUnavailableError as e:
            logger.error(f"[Service Unavailable Error] {e}")
            raise ReplyNeededError(message=f"[551721]Service Unavailable {e}")
        except RateLimitError as e:
            logger.error(f"ApiEndPoint:{e}")
            raise ReplyNeededError(message=f"[551580]Rate Limit Error {e}")
        except ProviderException as e:
            logger.info(f"[Service Provider]{e}")
            raise ReplyNeededError(message=f"[551183]Service Provider Error {e}")
        except PydanticValidationError as e:
            logger.exception(e)
            raise ReplyNeededError(message="[551684]Request Data ValidationError")
        except Exception as e:
            logger.exception(e)
            raise e

    async def _flash(self,
                     *,
                     task: TaskHeader,
                     llm: OpenaiMiddleware,
                     auto_write_back: bool = True,
                     intercept_function: bool = False,
                     retrieve_message: bool = False,
                     disable_function: bool = False
                     ):
        """
        函数池刷新
        :param intercept_function: 是否拦截函数调用转发到函数处理器
        :param retrieve_message: 是否检索消息
        :param task: 任务
        :param llm: Openai中间件
        :param auto_write_back: 是否自动写回
        :param disable_function: 是否禁用函数
        :return:
        """
        try:
            try:
                _llm_result = await self.llm_request(
                    llm_agent=llm,
                    auto_write_back=auto_write_back,
                    disable_function=disable_function,
                    retrieve_message=retrieve_message
                )
                get_message = _llm_result.default_message
                logger.debug(f"[x] LLM Message Sent \n--message {get_message}")
                if not isinstance(get_message, AssistantMessage):
                    raise ReplyNeededError("[55682]Request Result Not Valid, Must Be `AssistantMessage`")
            except Exception as e:
                if isinstance(e, ReplyNeededError):
                    await self.sender.error(
                        receiver=task.receiver,
                        text=get_request_error_message(str(e))
                    )
                raise e
            if intercept_function:
                if get_message.sign_function:
                    await self.sender.reply(
                        receiver=task.receiver,
                        message=[get_message],
                        reply_to_message=False
                    )
                    await self.sender.function(
                        receiver=task.receiver,
                        task=task,
                        llm=llm,  # IMPORTANT
                        llm_result=_llm_result
                    )
                    return logger.debug("Function loop ended")
            return await self.sender.reply(
                receiver=task.receiver,
                message=[get_message]
            )
        except Exception as e:
            raise e

    async def deal_message(self, message) -> Tuple[
        Optional[TaskHeader], Optional[OpenaiMiddleware], Optional[str], Optional[bool]
    ]:
        """
        处理消息
        """
        logger.trace(f"[x] Received Raw Task \n--message {message.body}")
        _task: TaskHeader = TaskHeader.model_validate_json(message.body.decode("utf-8"))
        # 没有任何参数
        if _task.task_meta.direct_reply:
            await self.sender.forward(
                receiver=_task.receiver,
                message=_task.message
            )
            return _task, None, "direct_reply", _task.task_meta.release_chain

        functions = await FunctionReorganize(task=_task).build_arg()
        """函数组建，自动过滤拉黑后的插件和错误过多的插件"""
        try:
            _llm = OpenaiMiddleware(
                task=_task,
                functions=functions,
                tools=[]
                # 内部会初始化函数工具，这里是其他类型工具
            ).init()
        except ProviderException as e:
            await self.sender.error(
                receiver=_task.receiver,
                text=f"🥞 Auth System Report {formatting.escape_markdown(str(e))}"
            )
            raise e
        """构建通信代理"""
        schema = _llm.get_schema()
        logger.debug(f"[x] Received Order \n--order {_task.model_dump_json()}")
        # function_response write back
        if _task.task_meta.write_back:
            for call in _task.task_meta.callback:
                if schema.func_executor == "tool_call":
                    _func_tool_msg = call.get_tool_message()
                elif schema.func_executor == "function_call":
                    _func_tool_msg = call.get_function_message()
                elif schema.func_executor == "unsupported":
                    _func_tool_msg = None
                else:
                    raise NotImplementedError(f"func_executor {schema.func_executor} not implemented")
                """消息类型是由请求结果决定的。也就是理论不存在预料外的冲突。"""

                _llm.write_back(
                    message=_func_tool_msg
                )
                logger.debug(f"[x] Function Response Write Back \n--callback {call.name}")

        # 插件直接转发与重处理
        if _task.task_meta.callback_forward:
            # 插件数据响应到前端
            if _task.task_meta.callback_forward_reprocess:
                # 手动写回则禁用从 Task 数据体自动回写
                # 防止AI去启动其他函数，禁用函数
                await self._flash(
                    llm=_llm,
                    task=_task,
                    intercept_function=True,
                    disable_function=True,
                    auto_write_back=False
                )
                # 同时递交部署点
                return _task, _llm, "callback_forward_reprocess", _task.task_meta.release_chain

            # 转发函数
            await self.sender.forward(
                receiver=_task.receiver,
                message=_task.message
            )
            # 同时递交部署点
            return _task, _llm, "callback_forward", _task.task_meta.release_chain

        await self._flash(llm=_llm, task=_task, intercept_function=True)
        return _task, None, "default", _task.task_meta.release_chain

    async def on_message(self, message: AbstractIncomingMessage):
        if not self.task or not self.sender:
            raise ValueError("receiver not set core")
        try:
            if os.getenv("LLMBOT_STOP_REPLY") == "1":
                return None
            # 处理消息
            task, llm, point, release = await self.deal_message(message)
            logger.debug(f"--point {point}")
            # 启动链式函数应答循环
            if release and task:
                chain: Chain = await ChainReloader(uid=task.receiver.uid).get_task()
                if chain:
                    await Task(queue=chain.channel).send_task(task=chain.arg)
                    logger.info(f"🧀 Chain point release\n--callback_send_by {point}")
                else:
                    logger.info(f"🧀🧀 Chain point release but empty {point}")
        except Exception as e:
            logger.exception(e)
            await message.reject(requeue=False)
        else:
            await message.ack(multiple=False)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/14 上午11:31
# @Author  : sudoskys
# @File    : test_chain_box.py
# @Software: PyCharm
import sys

sys.path.append("..")

from pprint import pprint  # noqa: E402

from llmkira.middleware import chain_box  # noqa: E402
from llmkira.task.schema import TaskHeader  # noqa: E402


def test_schema():
    arg = TaskHeader(sender=TaskHeader.Location(), receiver=TaskHeader.Location())
    print("\n")
    pprint(arg)
    try:
        chain_box.schema.Chain(
            thead_uuid="550w",
            arg=arg,
            channel="5slist",
            creator_uid="550w",
        )
    except Exception as e:
        pprint(f"As Expected {str(e)}")
    else:
        assert False, "Should Raise Error"
    test_obj = chain_box.schema.Chain(
        thead_uuid="550w",
        arg=arg,
        channel="5slist",
        creator_uid="platform:2225",
    )
    print("\n")
    pprint(test_obj)



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 上午11:47
# @Author  : sudoskys
# @File    : default_factory.py
# @Software: PyCharm
import re
from typing import Any

from . import resign_transfer
from .schema import Builder, Parser, TransferMata


@resign_transfer()
class DefaultMessageBuilder(Builder):
    sign = TransferMata(
        platform=re.compile(r".*"),  # 匹配所有
        plugin_name="default",
        agent="receiver",
        priority=0
    )

    async def pipe(self, arg) -> Any:
        return arg


@resign_transfer()
class DefaultMessageParser(Parser):
    sign = TransferMata(
        platform=re.compile(r".*"),  # 匹配所有
        plugin_name="default",
        agent="sender",
        priority=0
    )

    async def pipe(self, arg) -> Any:
        return arg



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 下午10:22
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import base64
import binascii
import json
import random
from typing import List, Tuple

import crescent
import hikari
from hikari import Intents
from hikari.impl import ProxySettings
from llmkira.extra.user import UserControl
from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.router import RouterManager, Router
from llmkira.sdk.func_calling import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.setting.discord import BotSetting
from llmkira.task import Task, TaskHeader
from loguru import logger
from telebot import formatting
from typing_extensions import Annotated

from .event import help_message
from ..schema import Runner

__sender__ = "discord_hikari"
__default_function_enable__ = True

from ..util_func import auth_reloader, is_command, is_empty_command
from ...error import get_upload_error_message
from ...sdk.openapi.trigger import get_trigger_loop
from ...sdk.schema import File

DiscordTask = Task(queue=__sender__)


class StartUpEvent(object):
    def __init__(self) -> None:
        ...

    async def on_start(self, event: hikari.StartedEvent) -> None:
        """
        This function is called when your bot starts. This is a good place to open a
        connection to a database, aiohttp client, or similar.
        """
        pass

    async def on_stop(self, event: hikari.StoppedEvent) -> None:
        """
        This function is called when your bot stops. This is a good place to put
        cleanup functions for the model class.
        """
        pass


class DiscordBotRunner(Runner):

    def __init__(self):
        self.bot = None
        self.proxy = None

    async def upload(self, attachment: hikari.Attachment, uid: str):
        # Limit 7MB
        if attachment.size > 1024 * 1024 * 7:
            raise Exception("File size too large")
        file_name = f"{attachment.filename}"
        file_data = await attachment.read()
        return await File.upload_file(file_name=file_name,
                                      file_data=file_data,
                                      creator_uid=uid
                                      )

    async def run(self):
        if not BotSetting.available:
            logger.warning("Sender Runtime:Discord not configured, skip")
            return None
        my_intents = (
                Intents.GUILDS |
                Intents.GUILD_MESSAGES |
                Intents.DM_MESSAGES |
                Intents.MESSAGE_CONTENT
        )
        logger.info(f"Sender Runtime:DiscordBot will start with intents:{my_intents}")
        self.bot = hikari.GatewayBot(
            intents=my_intents,
            token=BotSetting.token,
            proxy_settings=ProxySettings(
                url=BotSetting.proxy_address
            ) if BotSetting.proxy_address else None
        )
        self.proxy = BotSetting.proxy_address
        # prepare
        bot = self.bot
        model = StartUpEvent()
        client = crescent.Client(app=bot, model=model)
        # Base64 解码
        try:
            _based = BotSetting.token.split(".", maxsplit=1)[0] + "=="
            _bot_id = base64.b64decode(_based).decode("utf-8")
        except UnicodeDecodeError as e:
            logger.exception(f"Sender Runtime:DiscordBot token maybe invalid {e}")
        except binascii.Error as e:
            logger.exception(f"Sender Runtime:DiscordBot token maybe invalid {e}")
        except Exception as e:
            logger.exception(f"Sender Runtime:DiscordBot token maybe invalid {e}")
        else:
            BotSetting.bot_id = _bot_id

        # Task Creator
        async def create_task(message: hikari.Message, funtion_enable: bool = False):
            # event.message.embeds
            _file = []
            for attachment in message.attachments:
                try:
                    _file.append(
                        await self.upload(
                            attachment=attachment,
                            uid=UserControl.uid_make(__sender__, message.author.id)
                        )
                    )
                except Exception as e:
                    logger.exception(e)
                    await message.respond(
                        content=get_upload_error_message(filename=attachment.filename, error=str(e)),
                        mentions_reply=True
                    )
            if message.content:
                if message.content.startswith(("/chat", "/task")):
                    message.content = message.content[5:]
                if message.content.startswith("/ask"):
                    message.content = message.content[4:]
            message.content = message.content if message.content else ""
            _user_name = bot.get_me().username
            if _user_name:
                message.content = message.content.replace(f"<@{BotSetting.bot_id}>", f" @{_user_name} ")
            logger.info(
                f"discord_hikari:create task from {message.channel_id} "
                f"{message.content[:300]} funtion_enable:{funtion_enable}"
            )
            # 任务构建
            try:
                # 转析器
                message, _file = await self.loop_turn_only_message(
                    platform_name=__sender__,
                    message=message,
                    file_list=_file
                )
                # Reply
                success, logs = await DiscordTask.send_task(
                    task=TaskHeader.from_discord_hikari(
                        message,
                        file=_file,
                        deliver_back_message=[],
                        task_meta=TaskHeader.Meta.from_root(
                            function_enable=funtion_enable,
                            release_chain=True,
                            platform=__sender__
                        ),
                        trace_back_message=[]
                    )
                )
                if not success:
                    pass
            except Exception as e:
                logger.exception(e)

        async def endpoint_autocomplete(
                ctx: crescent.AutocompleteContext, option: hikari.AutocompleteInteractionOption
        ) -> List[Tuple[str, str]]:
            return [("https://api.openai.com/v1/chat/completions", "https://api.openai.com/v1/chat/completions")]

        @client.include
        @crescent.command(dm_enabled=True, name="token_clear", description="clear your service provider token")
        async def listen_token_clear_command(ctx: crescent.Context):
            try:
                status = "🪄 Clear token success"
                await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, ctx.user.id),
                    token=None
                )
            except Exception as e:
                status = "❌ Clear token failed"
                logger.error(f"[102335]token_clear failed {e}")
            return await ctx.respond(
                content=status,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="token", description="set your service provider token")
        async def listen_token_command(
                ctx: crescent.Context,
                token: str,
        ):
            try:
                token = await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, ctx.user.id),
                    token=token
                )
            except Exception as e:
                return await ctx.respond(
                    content=f"❌ Set token failed\n`{e}`",
                    ephemeral=True
                )
            else:
                return await ctx.respond(
                    content=formatting.format_text(
                        f"🪄 Set token success\n",
                        f"Set {token}"
                    ),
                    ephemeral=True
                )

        @client.include
        @crescent.command(dm_enabled=True, name="func_unban", description="Unban some function")
        async def listen_func_unban_command(
                ctx: crescent.Context,
                func_name: str
        ):
            try:

                ban_list = await UserControl.unblock_plugin(
                    uid=UserControl.uid_make(__sender__, ctx.user.id),
                    plugin_name=func_name
                )
                status = "🪄 Unban success" + f"\nCurrent ban list:{ban_list}"
            except Exception as e:
                status = f"❌ Unban failed:{e}"
                logger.error(f"[2185536]unban func failed {e}")
            return await ctx.respond(
                content=status,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="func_ban", description="set your service provider token")
        async def listen_func_ban_command(
                ctx: crescent.Context,
                func_name: str,
        ):
            try:
                func_list = await UserControl.block_plugin(
                    uid=UserControl.uid_make(__sender__, ctx.user.id),
                    plugin_name=func_name
                )
            except Exception as e:
                return await ctx.respond(
                    content=f"❌ Ban failed\n`{e}`",
                    ephemeral=True
                )
            else:
                return await ctx.respond(
                    content=formatting.format_text(
                        f"🪄 Ban success\n",
                        f"Ban {func_name}\n",
                        f"Current ban list:{func_list}"
                    ),
                    ephemeral=True
                )

        @client.include
        @crescent.command(dm_enabled=True, name="clear_endpoint", description="clear openai endpoint")
        async def listen_clear_endpoint_command(ctx: crescent.Context):
            try:
                status = "🪄 Clear endpoint success"
                await UserControl.clear_endpoint(uid=UserControl.uid_make(__sender__, ctx.user.id))
            except Exception as e:
                status = "❌ Clear endpoint failed"
                logger.error(f"[102335]clear_endpoint failed {e}")
            return await ctx.respond(
                content=status,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="set_endpoint", description="set openai endpoint for yourself")
        async def listen_endpoint_command(
                ctx: crescent.Context,
                endpoint: Annotated[str, crescent.Autocomplete[endpoint_autocomplete]],
                openai_key: str,
                model_name: str
        ):
            try:
                new_driver = await UserControl.set_endpoint(
                    uid=UserControl.uid_make(__sender__, ctx.user.id),
                    endpoint=endpoint,
                    api_key=openai_key,
                    model=model_name
                )
            except Exception as e:
                return await ctx.respond(
                    content=f"❌ Set endpoint failed\n`{e}`",
                    ephemeral=True
                )
            else:
                return await ctx.respond(
                    content=formatting.format_text(
                        f"🪄 Set endpoint success\n",
                        new_driver.detail
                    ),
                    ephemeral=True
                )

        @client.include
        @crescent.command(dm_enabled=True, name="bind", description="bind some platform")
        async def listen_bind_command(ctx: crescent.Context, token: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=ctx.user.id, dsn=token)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=ctx.user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await ctx.respond(
                    content=f"`{e}`",
                    ephemeral=True
                )
            return await ctx.respond(
                content=formatting.format_text(
                    formatting.mbold("🪄 Bind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="unbind", description="unbind some platform")
        async def listen_unbind_command(ctx: crescent.Context, token: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=ctx.user.id, dsn=token)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=ctx.user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await ctx.respond(
                    content=f"`{e}`",
                    ephemeral=True
                )
            return await ctx.respond(
                content=formatting.format_text(
                    formatting.mbold("🪄 Unbind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="clear", description="clear your message  history")
        async def listen_clear_command(ctx: crescent.Context):
            RedisChatMessageHistory(session_id=f"{__sender__}:{ctx.user.id}", ttl=60 * 60 * 1).clear()
            _comment = ["I swear I've forgotten about you.", "Okay?", "Let's hope so.", "I'm not sure what you mean.",
                        "what about u?"]
            return await ctx.respond(
                ephemeral=True,
                content=formatting.format_text(
                    f"I have cleared your message history\n",
                    random.choice(_comment),
                )
            )

        @client.include
        @crescent.command(dm_enabled=True, name="help", description="show help message")
        async def listen_help_command(ctx: crescent.Context):
            return await ctx.respond(
                ephemeral=True,
                content=formatting.format_text(
                    f"**🥕 Help**",
                    help_message(),
                )
            )

        @client.include
        @crescent.command(dm_enabled=True, name="auth", description="auth [uuid]")
        async def listen_auth_command(ctx: crescent.Context, uuid: str):
            try:
                await auth_reloader(uuid=uuid, user_id=f"{ctx.user.id}", platform=__sender__)
            except Exception as e:
                message = "❌ Auth failed,You dont have permission or the task do not exist"
                logger.error(f"[270563]auth_reloader failed {e}")
            else:
                message = "🪄 Auth Pass"
            return await ctx.respond(
                content=message,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="tool", description="Show function tool list")
        async def listen_tool_command(ctx: crescent.Context):
            _tool = ToolRegister().get_plugins_meta
            _paper = [[tool_item.name, tool_item.get_function_string, tool_item.usage] for tool_item in _tool]
            arg = [
                f"**{item[0]}**\n"
                + f"`{item[1]}`\n"
                + f"{item[2]}\n"
                for item in _paper
            ]
            reply_message_text = formatting.format_text(
                formatting.mbold("🔧 Tool List"),
                *arg,
                separator="\n"
            )
            await ctx.respond(
                ephemeral=True,
                content=reply_message_text,
            )

        @client.include
        @crescent.command(dm_enabled=True, name="env", description="env VAR1=XXX")
        async def listen_env_command(ctx: crescent.Context, env_string: str):
            _manager = EnvManager.from_meta(platform=__sender__, user_id=ctx.user.id)
            try:
                _meta_data = _manager.parse_env(env_string=env_string)
                updated_env = await _manager.update_env(env=_meta_data)
            except Exception as e:
                logger.exception(f"[1202359]env update failed {e}")
                text = formatting.format_text(
                    f"**🧊 Env parse failed...O_o**\n",
                    separator="\n"
                )
            else:
                text = formatting.format_text(
                    f"**🧊 Updated**\n"
                    f"```json\n{json.dumps(updated_env, indent=2)}```",
                    separator="\n"
                )
            await ctx.respond(
                ephemeral=True,
                content=text,
            )

        # Two input point
        @client.include
        @crescent.event
        async def on_guild_create(event_: hikari.GuildMessageCreateEvent):
            if event_.message.author.is_bot:
                return
            if not event_.content:
                logger.info(f"discord_hikari:ignore a empty message,do you turn on the MESSAGE_CONTENT setting?")
                return
            # 扳机
            trigger = await get_trigger_loop(
                platform_name=__sender__,
                message=event_.content,
                uid=UserControl.uid_make(__sender__, event_.message.author.id)
            )
            if trigger:
                if trigger.action == "allow":
                    return await create_task(event_.message, funtion_enable=trigger.function_enable)
                if trigger.action == "deny":
                    return await event_.message.respond(content=trigger.message)
            # 命令
            # Bot may cant read message
            if is_command(text=event_.content, command=f"{BotSetting.prefix}chat"):
                if is_empty_command(text=event_.content):
                    return await event_.message.respond(content="?")
                return await create_task(event_.message, funtion_enable=__default_function_enable__)

            if is_command(text=event_.content, command=f"{BotSetting.prefix}task"):
                if is_empty_command(text=event_.content):
                    return await event_.message.respond(content="?")
                return await create_task(event_.message, funtion_enable=True)

            if is_command(text=event_.content, command=f"{BotSetting.prefix}ask"):
                if is_empty_command(text=event_.content):
                    return await event_.message.respond(content="?")
                return await create_task(event_.message, funtion_enable=False)

            if f"<@{BotSetting.bot_id}>" in event_.content:
                # At 事件
                return await create_task(event_.message, funtion_enable=__default_function_enable__)

            if event_.message.referenced_message:
                # 回复了 Bot
                if event_.message.referenced_message.author.id == bot.get_me().id:
                    return await create_task(event_.message, funtion_enable=__default_function_enable__)

        @client.include
        @crescent.event
        async def on_dm_create(event_: hikari.DMMessageCreateEvent):
            if event_.message.author.is_bot:
                return
            # 扳机
            trigger = await get_trigger_loop(platform_name=__sender__, message=event_.content)
            if trigger:
                if trigger.action == "allow":
                    return await create_task(event_.message, funtion_enable=trigger.function_enable)
                if trigger.action == "deny":
                    return await event_.message.respond(content=trigger.message)
            # 命令
            if is_command(text=event_.content, command=f"{BotSetting.prefix}task"):
                return await create_task(event_.message, funtion_enable=True)
            if is_command(text=event_.content, command=f"{BotSetting.prefix}ask"):
                return await create_task(event_.message, funtion_enable=False)
            return await create_task(event_.message, funtion_enable=__default_function_enable__)

        logger.success("Sender Runtime:DiscordBot start")
        bot.run()



# -*- coding: utf-8 -*-
import os
import re
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from types import ModuleType
from typing import Optional, Type, Dict, Any, List, Union, Set, final, Literal
from typing import TYPE_CHECKING

from loguru import logger
from pydantic import field_validator, BaseModel, Field, model_validator, PrivateAttr

if TYPE_CHECKING:
    from ...task import TaskHeader
    from ..schema import Function, TaskBatch
    from .model import PluginManager


class BaseTool(ABC, BaseModel):
    """
    基础工具类，所有工具类都应该继承此类
    """

    __slots__ = ()
    silent: bool = Field(False, description="是否静默")
    function: "Function" = Field(..., description="功能")
    keywords: List[str] = Field([], description="关键词")
    pattern: Optional[re.Pattern] = Field(None, description="正则匹配")
    require_auth: bool = Field(False, description="是否需要授权")
    repeatable: bool = Field(False, description="是否可重复使用")
    deploy_child: Literal[0, 1] = Field(1, description="如果为0，终结于此链点，不再向下传递")
    require_auth_kwargs: dict = {}
    env_required: List[str] = Field([], description="环境变量要求,ALSO NEED env_prefix")
    env_prefix: str = Field("", description="环境变量前缀")
    file_match_required: Optional[re.Pattern] = Field(None, description="re.compile 文件名正则")
    extra_arg: Dict[Any, Any] = Field({}, description="额外参数")
    __run_arg: Dict[Any, Any] = PrivateAttr(default_factory=dict)

    # exp: re.compile(r"file_id=([a-z0-9]{8})")

    @final
    def get_os_env(self, env_name):
        """
        获取 PLUGIN_+ 公共环境变量
        """
        env = os.getenv("PLUGIN_" + env_name, None)
        return env

    @final
    @property
    def env_list(self):
        """
        #🍪# If you think its shouldnt be final, you can issue it.
        """
        return [f"{self.env_prefix}{i}" for i in self.env_required]

    @final
    @property
    def name(self):
        """
        工具名称
        """
        return self.function.name

    @final
    @model_validator(mode="after")
    def _check_conflict(self):
        if self.silent and self.env_required:
            # raise ValueError("silent and env_required can not be True at the same time")
            logger.warning(
                "\n Plugin `silent` And `env_required` Should Not Be True Same Time"
                "Please Validate `env_required` In Execute Manual, Or Provide Callback To User Instead"
            )
        if self.env_required and not self.env_prefix:
            raise ValueError("env_required must be used with env_prefix")
        return self

    @final
    @field_validator("keywords", mode="before")
    def _check_keywords(cls, v):
        for i in v:
            if not isinstance(i, str):
                raise ValueError(f"keyword must be str, got {type(i)}")
            if len(i) > 20:
                raise ValueError(f"keyword must be less than 20 characters, got {len(i)}")
            if len(i) < 2:
                raise ValueError(f"keyword must be more than 2 characters, got {len(i)}")
        return v

    @classmethod
    def env_help_docs(cls, empty_env: List[str]) -> str:
        """
        环境变量帮助文档
        :param empty_env: 未被配置的环境变量列表
        :return: 帮助文档/警告
        """
        assert isinstance(empty_env, list), "empty_env must be list"
        return "You need to configure ENV to start use this tool"

    @abstractmethod
    def pre_check(self) -> Union[bool, str]:
        """
        预检查，如果不合格则返回 False，合格则返回 True
        返回字符串表示不合格，且有原因
        """
        return ...

    @abstractmethod
    def func_message(self, message_text, **kwargs):
        """
        如果合格则返回message，否则返回None，表示不处理
        message_text: 消息文本
        message_raw: 消息原始数据 `RawMessage`
        address: 消息地址 `tuple(sender,receiver)`
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    @abstractmethod
    async def failed(self,
                     task: "TaskHeader", receiver: "TaskHeader.Location",
                     exception, env: dict,
                     arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                     ):
        """
        通常为 回写消息+通知消息
        :param task: 任务
        :param receiver: 接收者
        :param exception: 异常
        :param env: 环境变量
        :param arg: 参数
        :param pending_task: 任务批次
        :param refer_llm_result: 上一次的结果
        """
        return ...

    @abstractmethod
    async def callback(self,
                       task: "TaskHeader", receiver: "TaskHeader.Location",
                       env: dict,
                       arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None
                       ):
        """
        运行成功会调用此函数
        :param task: 任务
        :param receiver: 接收者
        :param arg: 参数
        :param env: 环境变量
        :param pending_task: 任务批次
        :param refer_llm_result: 上一次的结果
        """
        return ...

    @abstractmethod
    async def run(self, *,
                  task: "TaskHeader", receiver: "TaskHeader.Location",
                  arg: dict, env: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                  ):
        """
        处理函数并返回回写结果
        :param task: 任务
        :param receiver: 接收者
        :param arg: 参数
        :param env: 环境变量
        :param pending_task: 任务批次
        :param refer_llm_result: 上一次的结果
        """
        return ...

    @final
    async def load(self, *,
                   task: "TaskHeader", receiver: "TaskHeader.Location",
                   arg: dict, env: dict = None, pending_task: "TaskBatch", refer_llm_result: dict = None,
                   **kwargs
                   ) -> dict:
        """
        加载工具，执行前的准备工作
        :param task: 任务
        :param receiver: 接收者
        :param arg: 参数
        :param env: 环境变量
        :param pending_task: 任务批次
        :param refer_llm_result: 上一次的结果
        :param kwargs: 额外参数
        :return: 运行结果
        """
        if not env:
            env = {}
        # 拒绝循环引用, 复制一份传递
        self.__run_arg = {
            "task": task.model_copy(deep=True),
            "receiver": receiver.model_copy(deep=True),
            "arg": arg,
            "env": env,
            "pending_task": pending_task.model_copy(deep=True),
            "refer_llm_result": refer_llm_result
        }
        self.__run_arg.update(kwargs)
        try:
            run_result = await self.run(
                **self.__run_arg
            )
        except Exception as e:
            logger.exception(e)
            logger.error(f"ToolExecutor:run error: {e}")
            self.__run_arg["exception"] = str(e)
            await self.failed(
                **self.__run_arg
            )
        else:
            self.__run_arg["result"] = run_result
            try:
                await self.callback(
                    **self.__run_arg
                )
            except Exception as e:
                logger.error(f"ToolExecutor:callback error: {e}")
        return self.__run_arg


@dataclass(eq=False)
class FuncPair:
    function: "Function"
    tool: Type[BaseTool]
    extra_arg: Dict[Any, Any] = field(default_factory=dict)

    @property
    def name(self):
        return self.function.name


class PluginMetadata(BaseModel):
    """
    插件元信息，由插件编写者提供
    """

    name: str = Field(..., max_length=20)
    """插件名称"""
    description: str = Field(..., max_length=200)
    """插件功能介绍"""
    usage: str = Field(..., max_length=300)
    """插件使用方法"""
    openapi_version: str = Field(..., max_length=10)
    """适应版本"""
    function: Set[FuncPair] = Field(default_factory=set)
    """插件工具"""
    type: Optional[str] = None
    """插件类型"""
    homepage: Optional[str] = Field(None, max_length=50)
    """插件主页"""
    config: Optional[Type[BaseModel]] = None
    """插件配置项"""
    extra: Dict[Any, Any] = Field(default_factory=dict)
    """插件额外信息，可由插件编写者自由扩展定义"""

    @model_validator(mode="before")
    def limit(cls, values):
        if len(values.get("function", [])) == 0:
            raise ValueError("function can not be empty")
        if not values.get("openapi_version", None):
            raise ValueError("openapi_version can not be empty")
        if len(values.get("description", "")) > 200:
            raise ValueError("description must be less than 200 characters")
        if len(values.get("usage", "")) > 300:
            raise ValueError("usage must be less than 300 characters")
        return values

    @property
    def get_function_string(self) -> str:
        return "include{" + ",".join([f"{i.name}" for i in self.function]) + "}"


@dataclass(eq=False)
class Plugin:
    """
    机制用途，不由用户实例化
    """

    name: str
    """插件索引标识，Bot 使用 文件/文件夹 名称作为标识符"""
    module: ModuleType
    """插件模块对象"""
    module_name: str
    """点分割模块路径"""
    manager: "PluginManager"
    """导入该插件的插件管理器"""
    parent_plugin: Optional["Plugin"] = None
    """父插件"""
    sub_plugins: Set["Plugin"] = field(default_factory=set)
    """子插件集合"""
    metadata: Optional[PluginMetadata] = None
    """插件元信息"""

# 探针插件加载机制 参考 nonebot2 项目设计



# -*- coding: utf-8 -*-
# @Time    : 2023/10/20 下午5:33
# @Author  : sudoskys
# @File    : http_client.py
# @Software: PyCharm
import requests


class KookHttpClient(object):
    def __init__(self, token):
        self.base_url = 'https://www.kookapp.cn'
        self.bot_token = token

    def request(self, method, url, data=None):
        headers = {
            'Authorization': f'Bot {self.bot_token}'
        }
        response = requests.request(method, f'{self.base_url}{url}', headers=headers, json=data)
        return response.json()

    def create_channel_message(self, target_id, content, quote=None):
        data = {
            'type': 1,
            'target_id': target_id,
            'content': content,
            'quote': quote
        }
        return self.request('POST', '/api/v3/message/create', data)

    def create_direct_message(self, target_id, content, quote=None):
        data = {
            'type': 1,
            'target_id': target_id,
            'content': content,
            'quote': quote
        }
        return self.request('POST', '/api/v3/direct-message/create', data)



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午10:02
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from abc import abstractmethod, ABC

from llmkira.sdk.openapi.transducer import LoopRunner
from loguru import logger


class Runner(ABC):

    async def loop_turn_only_message(self, platform_name, message, file_list) -> tuple:
        """
        将 Openai 消息传入 Loop 进行修饰
        此过程将忽略掉其他属性。只留下 content
        """
        loop_runner = LoopRunner()
        trans_loop = loop_runner.get_sender_loop(platform_name=platform_name)
        await loop_runner.exec_loop(
            pipe=trans_loop,
            # sender Parser 约定的参数组合
            pipe_arg={
                "message": message,
                "files": file_list
            }
        )
        _arg = loop_runner.result_pipe_arg
        if not _arg.get("message"):
            logger.error("Message Loop Lose Message")
        new_message, new_file_list = _arg.get("message", message), _arg.get("files", file_list)
        return new_message, new_file_list

    @abstractmethod
    async def upload(self, *args, **kwargs):
        ...

    @abstractmethod
    def run(self, *args, **kwargs):
        ...



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 上午9:37
# @Author  : sudoskys
# @File    : llm_task.py
# @Software: PyCharm
import time
from typing import List, Optional

from loguru import logger
from pydantic import Field, BaseModel

from .llm_provider import GetAuthDriver
from ..extra.user import UserCost, CostControl
from ..schema import RawMessage, Scraper
from ..sdk.adapter import SCHEMA_GROUP
from ..sdk.endpoint import Driver
from ..sdk.endpoint.openai import Message
from ..sdk.endpoint.schema import LlmRequest, LlmResult
from ..sdk.memory.redis import RedisChatMessageHistory
from ..sdk.schema import ToolCallCompletion, SystemMessage, Function, Tool
from ..sdk.utils import sync
from ..task import TaskHeader


class SystemPrompt(BaseModel):
    """
    系统提示
    """
    start_tag: str = Field(default="[ACT CLAUSE]", description="开始标签")
    end_tag: str = Field(default="[CLAUSE END]", description="结束标签")
    content: List[str] = Field([], description="内容")

    def append(self, content: str):
        if isinstance(content, str):
            if len(content) > 3:
                self.content.append(f"* {content.upper()}")
        return self

    def clear(self):
        self.content = []

    @staticmethod
    def on_system() -> SystemMessage:
        top = SystemPrompt(
            start_tag="[ASSISTANT RULE]",
            end_tag="[RULE END]",
        )
        top.clear()
        top.append("DONT RE-USE THE FUNCTION WITH SAME PARAM")
        top.append("DONT REPEAT")
        top.append("REPLY IN SHORT-CONCISE")
        top.append("SPEAK IN MORE かわいい STYLE")
        top.append("ACT STEP BY STEP")
        top.append(f"<Time|{time.asctime(time.localtime(time.time()))}>")
        return top.prompt_message

    def prompt(self):
        if not self.content:
            return None
        return self.start_tag + "\n".join(self.content) + self.end_tag

    @property
    def prompt_message(self) -> Optional[SystemMessage]:
        """
        系统提示消息
        :raise: Exception if prompt message is empty
        :return: SystemMessage
        """
        content = self.prompt()
        if content:
            return SystemMessage(
                content=content
            )
        return None


class OpenaiMiddleware(object):
    """
    Openai中间件，用于处理消息转换和调用工具
    任务数据>转换器+函数填充>提取历史>放进刮削器>任务数据+刮削结果请求>获取Openai返回>进行声明通知/返回消息
    """

    def __init__(self,
                 task: TaskHeader,
                 functions: List[Function] = None,
                 tools: List[ToolCallCompletion] = None
                 ):
        self.auth_client = None
        self.driver = None
        if functions is None:
            functions = []
        if tools is None:
            tools = []

        assert isinstance(task, TaskHeader), "llm_task.py:task type error"

        self.functions: List[Function] = functions
        self._unique_function()
        # 必须先初始化函数
        self.tools: List[Tool] = tools
        self.scraper = Scraper()  # 刮削器
        self.task = task
        self.session_uid = task.receiver.uid
        self.system_prompt: SystemPrompt = SystemPrompt()
        self.message_history = RedisChatMessageHistory(
            session_id=self.session_uid,
            ttl=60 * 60 * 1
        )

    def init(self):
        """
        :raise: ProviderException
        """
        # 构建请求的驱动信息
        self.auth_client = GetAuthDriver(uid=self.session_uid)
        self.driver = sync(self.auth_client.get())
        assert isinstance(self.driver, Driver), "llm_task.py:GetAuthDriver s return not a driver!"
        return self

    def get_schema(self, model_name: str = None):
        if not model_name:
            model_name = self.driver.model
        return SCHEMA_GROUP.get_by_model_name(model_name=model_name)

    def _unique_function(self):
        """
        函数去重，禁止同名函数
        """
        _dict = {}
        for function in self.functions:
            assert isinstance(function, Function), "llm_task.py:function type error,cant unique"
            _dict[function.name] = function
        self.functions = list(_dict.values())

    def write_back(
            self,
            *,
            message: Optional[Message] = None
    ):
        """
        写回消息到 Redis 数据库中
        function 写回必须指定 name
        """
        if message:
            self.message_history.add_message(message=message)

    def _append_function_tools(self, functions: List[Function]):
        """

        """
        for function_item in functions:
            assert isinstance(function_item, Function), "llm_task.py:function type error"
            self.tools.append(
                Tool(
                    type="function",
                    function=function_item
                )
            )
        return self.tools

    def scraper_create_message(
            self,
            write_back=True,
            system_prompt=True
    ):
        """
        从人类消息和历史消息中构建请求所用消息
        :param write_back: 是否写回,如果是 False,那么就不会写回到 Redis 数据库中，也就是重新请求
        :param system_prompt: 是否添加系统提示
        :return: None
        """
        if system_prompt:
            # 此前和连带的消息都添加
            self.scraper.add_message(self.system_prompt.on_system(), score=1000)
            _plugin_system_prompt = self.system_prompt.prompt_message
            if _plugin_system_prompt is not None:
                assert isinstance(_plugin_system_prompt, Message), "llm_task.py:system prompt type error"
                self.scraper.add_message(_plugin_system_prompt, score=500)
        _history = []
        # database:read redis
        history_messages = self.message_history.messages
        for i, message in enumerate(history_messages):
            _history.append(message)
        # 刮削器合并消息，这里评价简写了。
        for i, _msg in enumerate(_history):
            self.scraper.add_message(_msg, score=len(str(_msg)))
        # 处理 人类 发送的消息
        if write_back:
            _buffer = []
            raw_message = self.task.message
            raw_message: List[RawMessage]
            for i, message in enumerate(raw_message):
                message: RawMessage
                # message covert
                _buffer.append(
                    message.format_user_message(
                        role="user",
                    )
                )
            # 装样子添加评分
            # TODO 评分机制
            for i, _msg in enumerate(_buffer):
                self.scraper.add_message(_msg, score=len(str(_msg)) + 50)
            # database:save redis
            for _msg in _buffer:
                self.message_history.add_message(message=_msg)

    async def request_openai(
            self,
            auto_write_back: bool,
            retrieve_mode: bool = False,
            disable_function: bool = False
    ) -> LlmResult:
        """
        处理消息转换和调用工具
        :param auto_write_back: 是否自动写回
        :param disable_function: 禁用函数
        :param retrieve_mode: 是否为检索模式，当我们需要重新处理超长消息时候，需要设定为 True
        :return: OpenaiResult
        """
        run_driver_model = self.driver.model if not retrieve_mode else self.driver.retrieve_model
        endpoint_schema = self.get_schema(model_name=run_driver_model)
        # 添加函数定义的系统提示
        if not disable_function:
            for function_item in self.functions:
                assert isinstance(function_item, Function), "llm_task.py:function type error"
                self.system_prompt.append(function_item.config.system_prompt)
        # 构建标准函数列表并转换添加到工具列表
        functions = [
            function_item.request_final(schema_model=run_driver_model)
            for function_item in self.functions
        ]
        self._append_function_tools(functions=functions)
        tools = [
            tool_item.request_final(schema_model=run_driver_model)
            for tool_item in self.tools
        ]
        # 构建消息列表
        self.scraper_create_message(
            write_back=auto_write_back
        )
        # 折叠消息列表
        if retrieve_mode:
            self.scraper.fold_message()
        # 削减消息列表
        self.scraper.reduce_messages(
            limit=endpoint_schema.token_limit,
            model_name=run_driver_model
        )
        # 获取消息列表
        messages = self.scraper.get_messages()
        # 标准化消息列表
        messages = [
            message.request_final(schema_model=run_driver_model)
            for message in messages
        ]
        # 日志
        logger.info(
            f"[x] Openai request"
            f"\n--message {messages} "
            f"\n--url {self.driver.endpoint} "
            f"\n--org {self.driver.org_id} "
            f"\n--model {run_driver_model} "
            f"\n--function {functions}"
            f"\n--retrieve_mode {retrieve_mode}"
        )
        # 禁用函数？
        if disable_function or not functions:
            logger.debug(f"Llm Driver --disable function:{disable_function}")
        # 必须校验
        if disable_function:
            functions = None
            tools = None
        # 根据模型选择不同的驱动a
        assert messages, "message cant be none"
        endpoint: LlmRequest = endpoint_schema.request.init(
            driver=self.driver,
            messages=messages,
            functions=functions,
            tools=tools,
            echo=False
        )
        # 调用Openai
        result: LlmResult = await endpoint.create()
        _message = result.default_message
        _usage = result.usage.total_tokens
        self.write_back(message=_message)
        # print(result.model_dump_json(indent=2))
        # 记录消耗
        await CostControl.add_cost(
            cost=UserCost.create_from_task(
                uid=self.session_uid,
                request_id=result.id,
                cost=UserCost.Cost(
                    cost_by=self.task.receiver.platform,
                    token_usage=_usage,
                    token_uuid=self.driver.uuid,
                    llm_model=self.driver.model,
                    provide_type=self.auth_client.provide_type().value
                )
            )
        )
        return result



# -*- coding: utf-8 -*-
# @Time    : 2023/10/21 下午6:25
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import ssl
from typing import List

from loguru import logger
from slack_sdk.web.async_client import AsyncWebClient

from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.receiver import function
from llmkira.receiver.receiver_client import BaseReceiver, BaseSender
from llmkira.receiver.slack.creat_message import ChatMessageCreator
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.schema import File, Message
from llmkira.setting.slack import BotSetting
from llmkira.task import Task, TaskHeader

__receiver__ = "slack"

from llmkira.middleware.router.schema import router_set

# 魔法
import nest_asyncio

nest_asyncio.apply()
# 设置路由系统
router_set(role="receiver", name=__receiver__)


class SlackSender(BaseSender):
    """
    平台路由
    """

    def __init__(self):
        if not BotSetting.available:
            return
        ssl_cert = ssl.SSLContext()
        if BotSetting.proxy_address:
            self.proxy = BotSetting.proxy_address
            logger.info("SlackBot proxy_tunnels being used in `AsyncWebClient`!")
        self.bot = AsyncWebClient(
            token=BotSetting.bot_token,
            ssl=ssl_cert,
            proxy=BotSetting.proxy_address
        )

    async def file_forward(self, receiver: TaskHeader.Location, file_list: List[File]):
        for file_obj in file_list:
            # URL FIRST
            if file_obj.file_url:
                await self.bot.files_upload_v2(
                    file=file_obj.file_url,
                    channels=receiver.chat_id,
                    thread_ts=receiver.message_id,
                    filename=file_obj.file_name,
                )
            # DATA
            _data: File.Data = await file_obj.raw_file()
            if not _data:
                logger.error(f"file download failed {file_obj.file_id}")
                continue
            try:
                await self.bot.files_upload_v2(
                    file=_data.file_data,
                    filename=_data.file_name,
                    channels=receiver.chat_id,
                    thread_ts=receiver.message_id,
                )
            except Exception as e:
                logger.error(e)
                logger.error(f"file upload failed {file_obj.file_id}")
                await self.bot.chat_postMessage(
                    channel=receiver.chat_id,
                    thread_ts=receiver.message_id,
                    text=str(f"Failed,Server down,or bot do not have *Bot Token Scopes* "
                             f"of `files:write` scope to upload file.")
                )

    async def forward(self, receiver: TaskHeader.Location, message: List[RawMessage]):
        """
        插件专用转发，是Task通用类型
        """
        for item in message:
            await self.file_forward(
                receiver=receiver,
                file_list=item.file
            )
            if item.only_send_file:
                continue
            _message = ChatMessageCreator(
                channel=receiver.chat_id,
                thread_ts=receiver.message_id
            ).update_content(message_text=item.text).get_message_payload(message_text=item.text)
            await self.bot.chat_postMessage(
                **_message
            )

    async def reply(self, receiver: TaskHeader.Location, message: List[Message], reply_to_message: bool = True):
        """
        模型直转发，Message是Openai的类型
        """
        for item in message:
            raw_message = await self.loop_turn_from_openai(platform_name=__receiver__, message=item, locate=receiver)
            await self.file_forward(
                receiver=receiver,
                file_list=raw_message.file
            )
            if raw_message.only_send_file:
                continue
            if not raw_message.text:
                continue
            assert raw_message.text, f"message content is empty"
            _message = ChatMessageCreator(
                channel=receiver.chat_id,
                thread_ts=receiver.message_id
            ).update_content(message_text=item.content).get_message_payload(message_text=item.content)
            await self.bot.chat_postMessage(
                channel=receiver.chat_id,
                thread_ts=receiver.message_id,
                text=raw_message.text
            )
        return logger.trace(f"reply message")

    async def error(self, receiver: TaskHeader.Location, text):
        _message = ChatMessageCreator(
            channel=receiver.chat_id,
            thread_ts=receiver.message_id
        ).update_content(message_text=text).get_message_payload(message_text=text)
        await self.bot.chat_postMessage(
            **_message
        )

    async def function(self,
                       receiver: TaskHeader.Location,
                       task: TaskHeader,
                       llm: OpenaiMiddleware,
                       llm_result: LlmResult,
                       ):
        task_batch = llm_result.default_message.get_executor_batch()
        verify_map = await self.push_task_create_message(
            task=task,
            task_batch=task_batch,
            llm_result=llm_result,
            receiver=receiver
        )
        new_receiver = task.receiver.model_copy()
        new_receiver.platform = __receiver__
        """更新接收者为当前平台，便于创建的函数消息能返回到正确的客户端"""
        new_meta = task.task_meta.pack_loop(
            plan_chain_pending=task_batch,
            verify_map=verify_map
        )
        """克隆元数据为当前平台"""

        await Task(queue=function.__receiver__).send_task(
            task=TaskHeader.from_function(
                llm_result=llm_result,
                task_meta=new_meta,
                receiver=new_receiver,
                message=task.message
            )
        )
        """发送打包后的任务数据"""
        del new_meta
        del task
        """清理内存"""


__sender__ = SlackSender()


class SlackReceiver(BaseReceiver):
    """
    receive message from telegram
    """

    async def slack(self):
        self.set_core(sender=__sender__, task=Task(queue=__receiver__))
        if not BotSetting.available:
            logger.warning("Receiver Runtime:Slack Setting empty")
            return None
        try:
            logger.success("Receiver Runtime:Slack start")
            await self.task.consuming_task(self.on_message)
        except KeyboardInterrupt:
            logger.warning("Slack Receiver shutdown")
        except Exception as e:
            logger.exception(e)
            raise e



# -*- coding: utf-8 -*-
# @Time    : 2023/10/21 下午2:09
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
import time
from ssl import SSLContext

import aiohttp
from llmkira.extra.user import UserControl
from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.router import RouterManager, Router
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.sender.util_func import is_command, auth_reloader, parse_command
from llmkira.setting.slack import BotSetting
from llmkira.task import Task, TaskHeader
from loguru import logger
from slack_bolt.adapter.socket_mode.async_handler import AsyncSocketModeHandler
from slack_bolt.app.async_app import AsyncApp
from slack_bolt.context.ack.async_ack import AsyncAck
from slack_bolt.context.respond.async_respond import AsyncRespond
from slack_sdk.web.async_client import AsyncWebClient
from telebot import formatting
from telebot.formatting import escape_markdown

from .event import SlashCommand, SlackChannelInfo, help_message
from .schema import SlackMessageEvent
from ..schema import Runner

__sender__ = "slack"

from ...sdk.openapi.trigger import get_trigger_loop
from ...sdk.schema import File

SlackTask = Task(queue=__sender__)
__default_function_enable__ = True
__join_cache__ = {}


async def download_url(url):
    async with aiohttp.ClientSession(
            headers={
                "Authorization": f"Bearer {BotSetting.bot_token}"
            } if url.startswith("https://files.slack.com") else None
    ) as session:
        async with session.get(url, allow_redirects=True) as response:
            # verify content type is html
            if response.content_type == "text/html":
                logger.error(f"Warning, may bot missing `file:read` scope! because content_type is text/html")
            if response.status == 200:
                return await response.read()


class SlackBotRunner(Runner):
    def __init__(self):
        self.proxy = None
        self.client = None
        self.bot = None

    async def upload(self, file: SlackMessageEvent.SlackFile, uid: str):
        if file.size > 1024 * 1024 * 20:
            return Exception(f"Chat File size too large:{file.size}")
        name = file.name
        url = file.url_private
        # Download from url
        try:
            data = await download_url(url=url)
        except Exception as e:
            logger.exception(f"[7652151]slack:download file failed :(\n {e} ,be sure you have the scope `files.read`")
            return Exception(f"Download file failed {e},be sure bot have the scope `files.read`")
        return await File.upload_file(file_name=name,
                                      file_data=data,
                                      creator_uid=uid
                                      )

    async def run(self):

        if not BotSetting.available:
            logger.warning("Sender Runtime:SlackBot not configured, skip")
            return
        ssl_cert = SSLContext()
        # pip3 install proxy.py
        # proxy --port 9000 --log-level d
        if BotSetting.proxy_address:
            self.proxy = BotSetting.proxy_address
            logger.info("SlackBot proxy_tunnels being used in `AsyncWebClient`!")
        self.client = AsyncWebClient(
            token=BotSetting.bot_token,
            ssl=ssl_cert,
            proxy=BotSetting.proxy_address
        )
        self.bot = AsyncApp(
            token=BotSetting.bot_token,
            signing_secret=BotSetting.secret,
            client=self.client,
        )
        bot = self.bot

        async def create_task(event_: SlackMessageEvent, funtion_enable: bool = False):
            """
            创建任务
            :param event_: SlackMessageEvent
            :param funtion_enable: 是否启用功能
            :return:
            """
            message = event_
            _file = []
            if message.text:
                if message.text.startswith(("/chat", "/task")):
                    message.text = message.text[5:]
                if message.text.startswith("/ask"):
                    message.text = message.text[4:]
                message.text = message.text.replace(f"<@{BotSetting.bot_id}>", f"@{BotSetting.bot_username}")
            if not message.text:
                return None
            for file in message.files:
                try:
                    _file.append(await self.upload(file=file, uid=UserControl.uid_make(__sender__, message.user)))
                except Exception as e:
                    await bot.client.chat_postMessage(
                        channel=message.channel,
                        text=formatting.format_text(
                            formatting.mbold("🪄 Failed"),
                            e,
                            separator="\n"
                        )
                    )
            message.text = message.text if message.text else ""
            logger.info(
                f"slack:create task from {message.channel} {message.text[:300]} funtion_enable:{funtion_enable}"
            )
            # 任务构建
            try:
                # 转析器
                message, _file = await self.loop_turn_only_message(
                    platform_name=__sender__,
                    message=message,
                    file_list=_file
                )
                # Reply
                success, logs = await SlackTask.send_task(
                    task=TaskHeader.from_slack(
                        message=message,
                        file=_file,
                        deliver_back_message=[],
                        task_meta=TaskHeader.Meta.from_root(
                            function_enable=funtion_enable,
                            release_chain=True,
                            platform=__sender__
                        )
                    )
                )
                if not success:
                    await bot.client.chat_postMessage(
                        channel=message.channel,
                        text=formatting.format_text(
                            formatting.mbold("🪄 Failed"),
                            logs,
                            separator="\n"
                        )
                    )
            except Exception as e:
                logger.exception(e)

        @bot.command(command='/clear_endpoint')
        async def listen_clear_endpoint_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            # _cmd, _arg = parse_command(command=message.text)
            _tips = "🪄 Done"
            await UserControl.clear_endpoint(uid=UserControl.uid_make(__sender__, command.user_id))
            return await respond(
                text=_tips
            )

        @bot.command(command='/set_endpoint')
        async def listen_set_endpoint_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            _except = _arg.split("#", maxsplit=2)
            if len(_except) == 3:
                openai_key, openai_endpoint, model = _except
            elif len(_except) == 2:
                openai_key, openai_endpoint = _except
                model = None
            else:
                openai_key, openai_endpoint = (_arg, None)
                model = None
            try:
                new_driver = await UserControl.set_endpoint(
                    uid=UserControl.uid_make(__sender__, command.user_id),
                    api_key=openai_key,
                    endpoint=openai_endpoint,
                    model=model
                )
            except Exception as e:
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(f"🪄 Failed: {e}", escape=False),
                        formatting.mitalic("Format: /set_endpoint <openai_key>#<openai_endpoint>#<llm_model>"),
                        formatting.mitalic(f"Model Name: {UserControl.get_model()}"),
                        separator="\n"
                    )
                )
            else:
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold("🪄 Done", escape=False),
                        new_driver.detail,
                        separator="\n"
                    )
                )

        @bot.command(command='/func_ban')
        async def listen_func_ban_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            try:
                func_list = await UserControl.block_plugin(
                    uid=UserControl.uid_make(__sender__, command.user_id),
                    plugin_name=_arg
                )
            except Exception as e:
                logger.error(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e), escape=False),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done", escape=False),
                    formatting.mitalic("Function Ban List"),
                    *[f"`{escape_markdown(item)}`" for item in func_list],
                    separator="\n"
                )
            )

        @bot.command(command='/func_unban')
        async def listen_func_unban_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            try:
                func_list = await UserControl.unblock_plugin(
                    uid=UserControl.uid_make(__sender__, command.user_id),
                    plugin_name=_arg
                )
            except Exception as e:
                logger.exception(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e), escape=False),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done", escape=False),
                    formatting.mitalic("Function Ban List"),
                    *[f"`{escape_markdown(item)}`" for item in func_list],
                    separator="\n"
                )
            )

        @bot.command(command='/token')
        async def listen_token_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            try:
                token = await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, command.user_id),
                    token=_arg
                )
            except Exception as e:
                logger.exception(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e), escape=False),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done", escape=False),
                    formatting.mcode(f"Bind Success {token}", escape=False),
                    separator="\n"
                )
            )

        @bot.command(command='/token_clear')
        async def listen_unbind_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            try:
                token = await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, command.user_id),
                    token=None
                )
            except Exception as e:
                logger.exception(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    formatting.mcode(f"Unbind Success {token}"),
                    separator="\n"
                )
            )

        @bot.command(command='/bind')
        async def listen_bind_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=command.user_id, dsn=_arg)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=command.user_id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e), escape=False),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done", escape=False),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                )
            )

        @bot.command(command='/unbind')
        async def listen_unbind_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=command.user_id, dsn=_arg)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=command.user_id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await respond(
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    )
                )
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                )
            )

        @bot.command(command='/env')
        async def listen_env_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return
            _arg = command.text
            _manager = EnvManager.from_meta(platform=__sender__, user_id=command.user_id)
            try:
                _meta_data = _manager.parse_env(env_string=_arg)
                updated_env = await _manager.update_env(env=_meta_data)
            except Exception as e:
                logger.exception(f"[213562]env update failed {e}")
                text = formatting.format_text(
                    formatting.mbold("🧊 Failed"),
                    separator="\n"
                )
            else:
                text = formatting.format_text(
                    formatting.mbold("🦴 Env Changed"),
                    formatting.mcode(json.dumps(updated_env, indent=2)),
                    separator="\n"
                )
            await respond(
                text=text
            )

        @bot.command(command='/clear')
        async def listen_clear_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            RedisChatMessageHistory(session_id=f"{__sender__}:{command.user_id}", ttl=60 * 60 * 1).clear()
            return await respond(
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    separator="\n"
                )
            )

        @bot.command(command='/help')
        async def listen_help_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            await respond(
                text=formatting.format_text(
                    formatting.mbold("🥕 Help"),
                    help_message(),
                    separator="\n"
                )
            )

        @bot.command(command='/tool')
        async def listen_tool_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            _tool = ToolRegister().get_plugins_meta
            _paper = [[tool_item.name, tool_item.get_function_string, tool_item.usage] for tool_item in _tool]
            arg = [
                formatting.mbold(item[0], escape=False)
                + "\n"
                + formatting.mcode(item[1])
                + "\n"
                + formatting.mitalic(item[2], escape=False)
                + "\n"
                for item in _paper
            ]
            reply_message_text = formatting.format_text(
                formatting.mbold("🔧 Tool List"),
                *arg,
                separator="\n"
            )
            return await respond(
                text=reply_message_text
            )

        async def auth_chain(uuid, user_id):
            try:
                await auth_reloader(uuid=uuid, user_id=f"{user_id}", platform=__sender__)
            except Exception as e:
                auth_result = "❌ Auth failed,You dont have permission or the task do not exist"
                logger.info(f"[3031]auth_reloader failed {e}")
            else:
                auth_result = "🪄 Auth Pass"
            return auth_result

        @bot.command(command='/auth')
        async def listen_auth_command(ack: AsyncAck, respond: AsyncRespond, command):
            command: SlashCommand = SlashCommand.model_validate(command)
            await ack()
            if not command.text:
                return await respond(
                    text="🥕 Please input uuid"
                )
            _arg = command.text.strip("`")
            auth_result = await auth_chain(uuid=_arg, user_id=command.user_id)
            return await respond(
                text=auth_result
            )

        async def validate_join(event_: SlackMessageEvent):
            """
            When Start, **validate every channel first event**
            """
            if __join_cache__.get(event_.channel):
                return True
            try:
                _res = await self.bot.client.conversations_info(
                    channel=event_.channel
                )
                _channel: SlackChannelInfo = SlackChannelInfo.model_validate(_res.get("channel", {}))
                if not _channel.is_member:
                    raise Exception("Not in channel")
            except Exception as e:
                logger.error(f"[353688]slack:validate_join failed {e}")
                try:
                    await self.bot.client.chat_postMessage(
                        channel=event_.user,
                        text="🥕 Please... invite me to the channel first :)"
                        if not event_.channel_type == "im" else "🥺 I cant reach u, call from u own chat... ",
                    )
                except Exception as e:
                    logger.error(f"[364258]slack:validate_join failed {e}")
                return False
            else:
                __join_cache__[event_.channel] = True
                return True

        async def deal_group(event_: SlackMessageEvent):
            """
            自动响应群组消息
            """
            if not event_.text:
                return None
            if not await validate_join(event_=event_):
                return None
            _text = event_.text
            # 扳机
            trigger = await get_trigger_loop(
                platform_name=__sender__,
                message=_text,
                uid=UserControl.uid_make(__sender__, event_.user)
            )
            if trigger:
                if trigger.action == "allow":
                    return await create_task(event_, funtion_enable=trigger.function_enable)
                if trigger.action == "deny":
                    return await bot.client.chat_postMessage(
                        channel=event_.channel,
                        text=trigger.message,
                        thread_ts=event_.thread_ts,
                    )
            # 默认指令
            if is_command(text=_text, command="!chat"):
                return await create_task(event_, funtion_enable=__default_function_enable__)
            if is_command(text=_text, command="!task"):
                return await create_task(event_, funtion_enable=True)
            if is_command(text=_text, command="!ask"):
                return await create_task(event_, funtion_enable=False)
            if is_command(text=_text, command="!auth") or is_command(text=_text, command="`!auth"):
                _cmd, _arg = parse_command(command=_text.strip("`"))
                if not _arg:
                    return None
                _arg = str(_arg).strip("`")
                auth_result = await auth_chain(uuid=_arg, user_id=event_.user)
                return await bot.client.chat_postMessage(
                    channel=event_.channel,
                    text=auth_result,
                    thread_ts=event_.thread_ts,
                )
            if f"<@{BotSetting.bot_id}>" in _text or _text.endswith(f"<@{BotSetting.bot_id}>"):
                return await create_task(event_, funtion_enable=__default_function_enable__)

        @bot.event("message")
        async def listen_im(_event, _logger):
            """
            自动响应私聊消息
            """

            _logger.info(event)
            event_: SlackMessageEvent = SlackMessageEvent.model_validate(_event)
            # 校验消息是否过期
            if int(float(event_.event_ts)) < int(time.time()) - 60 * 60 * 5:
                _logger.warning(f"slack: message expired {event_.event_ts}")
                return
            if not event_.text:
                return None
            if event_.channel_type == "im":
                return await deal_group(event_)
            if event_.channel_type == "group":
                return await deal_group(event_)
            if event_.channel_type == "channel":
                return await deal_group(event_)

        _self_get = await self.client.auth_test()
        logger.success(f"SlackBot init, bot_id:{_self_get}")
        BotSetting.bot_id = _self_get["user_id"]
        BotSetting.bot_username = _self_get["user"]
        logger.success("Sender Runtime:SlackBot start")
        listen_world = AsyncSocketModeHandler(bot, BotSetting.app_token)
        await listen_world.start_async()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 下午10:23
# @Author  : sudoskys
# @File    : discord.py
# @Software: PyCharm
from typing import Optional

from dotenv import load_dotenv
from loguru import logger
from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class DiscordBot(BaseSettings):
    """
    代理设置
    """
    token: Optional[str] = Field(None, validation_alias='DISCORD_BOT_TOKEN', strict=True)
    prefix: Optional[str] = Field("/", validation_alias="DISCORD_BOT_PREFIX")
    proxy_address: Optional[str] = Field(None, validation_alias="DISCORD_BOT_PROXY_ADDRESS")  # "all://127.0.0.1:7890"
    bot_id: Optional[str] = Field(None)
    model_config = SettingsConfigDict(env_file='.env', env_file_encoding='utf-8', extra="ignore")

    @model_validator(mode='after')
    def bot_setting_validator(self):
        if self.token is None:
            logger.info(f"🍀Discord Bot Token Not Set")
        if self.proxy_address:
            logger.success(f"DiscordBot proxy was set to {self.proxy_address}")
        return self

    @property
    def available(self):
        return self.token is not None


load_dotenv()
BotSetting = DiscordBot()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/20 下午5:33
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
import random

import aiohttp
from khl import Bot, Message, Cert, MessageTypes, PrivateMessage, PublicMessage
from llmkira.extra.user import UserControl
from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.router import RouterManager, Router
from llmkira.sdk.func_calling import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.setting.kook import BotSetting
from llmkira.task import Task, TaskHeader
from loguru import logger
from telebot import formatting

from .event import help_message, _upload_error_message_template, MappingDefault
from ..schema import Runner

__sender__ = "kook"
__default_function_enable__ = True

from ..util_func import auth_reloader, is_command, is_empty_command
from ...sdk.openapi.trigger import get_trigger_loop
from ...sdk.schema import File

KookTask = Task(queue=__sender__)


class CacheTemp(object):
    def __init__(self):
        self.__file_cache_queue_ = {}

    def new(self, user_id):
        if not self.__file_cache_queue_.get(user_id) or not isinstance(self.__file_cache_queue_[user_id], list):
            self.__file_cache_queue_[user_id] = []
        return self

    def append(self, user_id, item):
        self.new(user_id=user_id).__file_cache_queue_[user_id].append(item)
        return self

    def clear(self, user_id):
        self.new(user_id=user_id).__file_cache_queue_[user_id].clear()
        return self

    def get(self, user_id):
        return self.__file_cache_queue_.get(user_id, [])


_file_cache_queue_ = CacheTemp()


async def download_url(url):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status == 200:
                return await response.read()


class KookBotRunner(Runner):

    def __init__(self):
        self.bot = None
        self.proxy = None

    async def upload(self,
                     file_info: dict,
                     uid: str
                     ):
        if not file_info.get("url"):
            return Exception("File url not found")
        if file_info.get("type") == "file":
            if file_info.get("size") > 1024 * 1024 * 5:
                raise Exception("File size too large")
        name = file_info.get("name", "unknown")
        url = file_info.get("url", None)
        # Download from url
        try:
            data = await download_url(url=url)
        except Exception as e:
            logger.exception(f"[602253]kook:download file failed {e}")
            return Exception(f"Download file failed {e}")
        return await File.upload_file(file_name=name,
                                      file_data=data,
                                      creator_uid=uid
                                      )

    async def run(self):
        if not BotSetting.available:
            logger.warning("Sender Runtime:Kook not configured, skip")
            return None
        self.bot = Bot(cert=Cert(token=BotSetting.token))
        # prepare
        bot = self.bot

        # Task Creator
        async def create_task(_event: Message, funtion_enable: bool = False):
            # event.message.embeds
            _file: list = []
            try:
                if _event.type == MessageTypes.FILE:
                    _file_cache_queue_.append(
                        user_id=_event.author_id,
                        item=await self.upload(_event.extra.get("attachments"),
                                               uid=UserControl.uid_make(__sender__, _event.author_id))
                    )
                    return None
                if _event.type == MessageTypes.IMG:
                    _file_cache_queue_.append(
                        user_id=_event.author_id,
                        item=await self.upload(_event.extra.get("attachments"),
                                               uid=UserControl.uid_make(__sender__, _event.author_id))
                    )
                    return None
            except Exception as e:
                logger.exception(e)
                _template: str = random.choice(_upload_error_message_template)
                await _event.reply(
                    content=_template.format_map(map=MappingDefault(filename="File", error=str(e))),
                    type=MessageTypes.KMD,
                )
                return None

            # Cache Run Point
            if _event.type in [MessageTypes.KMD, MessageTypes.TEXT]:
                _file: list = [item for item in _file_cache_queue_.get(user_id=_event.author_id) if item]
                _file_cache_queue_.clear(user_id=_event.author_id)
            message: Message = _event
            if message.content:
                if message.content.startswith(("/chat", "/task")):
                    message.content = message.content[5:]
                if message.content.startswith("/ask"):
                    message.content = message.content[4:]
            message.content = message.content if message.content else ""
            logger.info(
                f"kook:create task from {message.ctx.channel.id} "
                f"{message.content[:300]} funtion_enable:{funtion_enable}"
            )
            # 任务构建
            try:
                # 转析器
                message, _file = await self.loop_turn_only_message(
                    platform_name=__sender__,
                    message=message,
                    file_list=_file
                )
                # Reply
                kook_task = TaskHeader.from_kook(
                    message,
                    file=_file,
                    deliver_back_message=[],
                    task_meta=TaskHeader.Meta.from_root(
                        function_enable=funtion_enable,
                        release_chain=True,
                        platform=__sender__
                    ),
                    trace_back_message=[]
                )
                success, logs = await KookTask.send_task(
                    task=kook_task
                )
                if not success:
                    pass
            except Exception as e:
                logger.exception(e)

        @bot.command(name='clear_endpoint')
        async def listen_clear_endpoint_command(msg: Message):
            try:
                status = "🪄 Clear endpoint success"
                await UserControl.clear_endpoint(uid=UserControl.uid_make(__sender__, msg.author_id))
            except Exception as e:
                status = "❌ Clear endpoint failed"
                logger.error(f"[102335]clear_endpoint failed {e}")
            if isinstance(msg, PublicMessage):
                return await msg.reply(
                    content=status,
                    is_temp=True
                )
            if isinstance(msg, PrivateMessage):
                return await msg.reply(
                    content=status
                )

        @bot.command(name='set_endpoint')
        async def listen_endpoint_command(
                msg: Message,
                openai_endpoint: str,
                openai_key: str,
                model: str
        ):
            try:
                new_driver = await UserControl.set_endpoint(
                    uid=UserControl.uid_make(__sender__, msg.author_id),
                    api_key=openai_key,
                    endpoint=openai_endpoint,
                    model=model
                )
            except Exception as e:
                return await msg.reply(
                    content=f"❌ Set endpoint failed\n`{e}`",
                    is_temp=True,
                    type=MessageTypes.KMD
                )
            else:
                return await msg.reply(
                    content=formatting.format_text(
                        f"🪄 Set endpoint success\n",
                        new_driver.detail
                    ),
                    is_temp=True,
                    type=MessageTypes.KMD
                )

        @bot.command(name='token_clear')
        async def listen_token_clear_command(msg: Message):
            try:
                status = "🪄 Clear token success"
                await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, msg.author_id),
                    token=None
                )
            except Exception as e:
                status = "❌ Clear token failed"
                logger.error(f"[217835]token clear failed {e}")
            if isinstance(msg, PublicMessage):
                return await msg.reply(
                    content=status,
                    is_temp=True
                )
            if isinstance(msg, PrivateMessage):
                return await msg.reply(
                    content=status
                )

        @bot.command(name='token')
        async def listen_token_command(
                msg: Message,
                token: str
        ):
            try:
                token = await UserControl.set_token(
                    uid=UserControl.uid_make(__sender__, msg.author_id),
                    token=token
                )
            except Exception as e:
                return await msg.reply(
                    content=f"❌ Set token failed\n`{e}`",
                    is_temp=True,
                    type=MessageTypes.KMD
                )
            else:
                return await msg.reply(
                    content=formatting.format_text(
                        f"🪄 Set token success\n",
                        token
                    ),
                    is_temp=True,
                    type=MessageTypes.KMD
                )

        @bot.command(name='func_ban')
        async def listen_func_ban_command(
                msg: Message,
                func_name: str
        ):
            try:
                func_list = await UserControl.block_plugin(
                    uid=UserControl.uid_make(__sender__, msg.author_id),
                    plugin_name=func_name
                )
            except Exception as e:
                return await msg.reply(
                    content=f"❌ Ban failed\n`{e}`",
                    is_temp=True,
                    type=MessageTypes.KMD
                )
            else:
                return await msg.reply(
                    content=formatting.format_text(
                        f"🪄 Ban success\n",
                        f"**🔗 Current Ban**\n"
                        f"{[f'`{item}`' for item in func_list]}"
                    ),
                    is_temp=True,
                    type=MessageTypes.KMD
                )

        @bot.command(name='func_unban')
        async def listen_func_unban_command(
                msg: Message,
                func_name: str
        ):
            try:
                func_list = await UserControl.unblock_plugin(
                    uid=UserControl.uid_make(__sender__, msg.author_id),
                    plugin_name=func_name
                )
            except Exception as e:
                return await msg.reply(
                    content=f"❌ Unban failed\n`{e}`",
                    is_temp=True,
                    type=MessageTypes.KMD
                )
            else:
                return await msg.reply(
                    content=formatting.format_text(
                        f"🪄 Ban success\n",
                        f"**🔗 Current Ban**\n"
                        f"{[f'`{item}`' for item in func_list]}"
                    ),
                    is_temp=True,
                    type=MessageTypes.KMD
                )

        @bot.command(name='bind')
        async def listen_bind_command(msg: Message, dsn: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=msg.author_id, dsn=dsn)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=msg.author_id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await msg.reply(
                    content=f"`{e}`",
                    is_temp=True,
                    type=MessageTypes.KMD
                )
            return await msg.reply(
                content=formatting.format_text(
                    formatting.mbold("🪄 Bind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                is_temp=True,
                type=MessageTypes.KMD
            )

        @bot.command(name='unbind')
        async def listen_unbind_command(msg: Message, dsn: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver_channel=__sender__, user_id=msg.author_id, dsn=dsn)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=msg.author_id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await msg.reply(
                    content=f"`{e}`",
                    type=MessageTypes.KMD,
                    is_temp=True
                )
            return await msg.reply(
                content=formatting.format_text(
                    formatting.mbold("🪄 Unbind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                type=MessageTypes.KMD,
                is_temp=True
            )

        @bot.command(name='clear')
        async def listen_clear_command(msg: Message):
            RedisChatMessageHistory(session_id=f"{__sender__}:{msg.author_id}", ttl=60 * 60 * 1).clear()
            _comment = ["I swear I've forgotten about you.", "Okay?", "Let's hope so.", "I'm not sure what you mean.",
                        "what about u?"]
            return await msg.reply(
                is_temp=True,
                type=MessageTypes.KMD,
                content=formatting.format_text(
                    f"I have cleared your message history\n",
                    random.choice(_comment),
                )
            )

        @bot.command(name='help')
        async def listen_help_command(msg: Message):
            return await msg.reply(
                is_temp=True,
                type=MessageTypes.KMD,
                content=formatting.format_text(
                    f"**🥕 Help**",
                    help_message(),
                )
            )

        @bot.command(name='tool')
        async def listen_tool_command(msg: Message):
            _tool = ToolRegister().get_plugins_meta
            _paper = [[tool_item.name, tool_item.get_function_string, tool_item.usage] for tool_item in _tool]
            arg = [
                formatting.mbold(item[0])
                + "\n"
                + f"`{item[1]}`"
                + "\n"
                + formatting.mitalic(item[2])
                + "\n"
                for item in _paper
            ]
            reply_message_text = formatting.format_text(
                formatting.mbold("🔧 Tool List"),
                *arg,
                separator="\n"
            )
            await msg.reply(
                is_temp=True,
                type=MessageTypes.KMD,
                content=reply_message_text,
            )

        @bot.command(name='auth')
        async def listen_auth_command(msg: Message, uuid: str):
            try:
                await auth_reloader(uuid=uuid, user_id=f"{msg.author_id}", platform=__sender__)
            except Exception as e:
                message = "❌ Auth failed,You dont have permission or the task do not exist"
                logger.error(f"[2753383]auth_reloader failed:{e}")
            else:
                message = "🪄 Auth Pass"
            return await msg.reply(
                content=message,
                is_temp=True,
                type=MessageTypes.KMD
            )

        @bot.command(name='env')
        async def listen_env_command(msg: Message, env_string: str):
            _manager = EnvManager.from_meta(platform=__sender__, user_id=msg.author_id)
            try:
                _meta_data = _manager.parse_env(env_string=env_string)
                updated_env = await _manager.update_env(env=_meta_data)
            except Exception as e:
                logger.exception(f"[1202359]env update failed {e}")
                text = formatting.format_text(
                    f"**🧊 Env parse failed...O_o**\n",
                    separator="\n"
                )
            else:
                text = formatting.format_text(
                    f"**🧊 Updated**\n"
                    f"```json\n{json.dumps(updated_env, indent=2)}```",
                    separator="\n"
                )
            await msg.reply(
                is_temp=True,
                type=MessageTypes.KMD,
                content=text,
            )

        async def on_guild_create(msg: PublicMessage):
            # 扳机
            trigger = await get_trigger_loop(
                platform_name=__sender__,
                message=msg.content,
                uid=UserControl.uid_make(__sender__, msg.author_id)
            )
            if trigger:
                if trigger.action == "allow":
                    return await create_task(msg, funtion_enable=trigger.function_enable)
                if trigger.action == "deny":
                    return await msg.reply(content=trigger.message)
            # 命令
            if is_command(text=msg.content, command=f"/chat"):
                if is_empty_command(text=msg.content):
                    return await msg.reply(content="?")
                return await create_task(msg, funtion_enable=__default_function_enable__)
            if is_command(text=msg.content, command=f"/chat"):
                if is_empty_command(text=msg.content):
                    return await msg.reply(content="?")
                return await create_task(msg, funtion_enable=True)
            if is_command(text=msg.content, command=f"/ask"):
                if is_empty_command(text=msg.content):
                    return await msg.reply(content="?")
                return await create_task(msg, funtion_enable=False)
            # 追溯回复

        async def on_dm_create(msg: PrivateMessage):
            # 扳机
            trigger = await get_trigger_loop(
                platform_name=__sender__,
                message=msg.content,
                uid=UserControl.uid_make(__sender__, msg.author_id)
            )
            if trigger:
                if trigger.action == "allow":
                    return await create_task(msg, funtion_enable=trigger.function_enable)
                if trigger.action == "deny":
                    return await msg.reply(content=trigger.message)
            if is_command(text=msg.content, command="/task"):
                return await create_task(msg, funtion_enable=True)
            if is_command(text=msg.content, command="/ask"):
                return await create_task(msg, funtion_enable=False)
            return await create_task(msg, funtion_enable=__default_function_enable__)

        @bot.command(regex=r'[\s\S]*')
        async def handle_message(event_: Message):
            if event_.author.bot:
                return None
            # NOT A BOT
            if isinstance(event_, PublicMessage):
                await on_guild_create(event_)
            if isinstance(event_, PrivateMessage):
                await on_dm_create(event_)

        logger.success("Sender Runtime:KookBot start")
        bot.run()



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午8:34
# @Author  : sudoskys
# @File    : aps.py
# @Software: PyCharm
from pathlib import Path

from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from loguru import logger

"""
from apscheduler.jobstores.redis import RedisJobStore
from redis.connection import ConnectionPool
import os
redis_url = os.getenv('REDIS_DSN', 'redis://localhost:6379/0')
job_store = RedisJobStore(
    connection_pool=ConnectionPool().from_url(redis_url)
)
"""
Path("config_dir").mkdir(exist_ok=True)
jobstores = {
    'default': SQLAlchemyJobStore(url='sqlite:///config_dir/aps.db')
}
executors = {
    'default': ThreadPoolExecutor(20)
}
job_defaults = {
    'coalesce': False,
    'max_instances': 3
}
from pytz import utc

SCHEDULER = AsyncIOScheduler(job_defaults=job_defaults,
                             timezone=utc,
                             executors=executors,
                             jobstores=jobstores
                             )


async def aps_start():
    logger.success("Receiver Runtime:APS Timer start")
    SCHEDULER.start()



# -*- coding: utf-8 -*-
# @Time    : 2023/11/14 上午10:09
# @Author  : sudoskys
# @File    : test_external.py
# @Software: PyCharm
import sys

sys.path.append("..")

from pydantic import ValidationError

from llmkira.external.schema import PluginExternal, PluginTestResultExport


def test_install():
    try:
        _install = PluginExternal.Install()
    except ValidationError as e:
        print("Cant init from empty")
    _install = PluginExternal.Install(pypi="test")
    _install = PluginExternal.Install(github="https://github.com/llmkira/")
    _install = PluginExternal.Install(shell="pip install llmkira")


def test_schema():
    _install = PluginExternal.Install(
        shell="pip install llmkira"
    )
    _external = PluginExternal(
        author_id="test",
        plugin_name="test.test.plugin",
        plugin_link="https://test.com",
        plugin_desc="desc",
        plugin_functions=[
            "test_func"
        ],
        org_id="test_org",
        plugin_install=_install
    )
    _result = PluginTestResultExport(plugin=_external, test_pass=True, out_date=True)
    assert _result
